import os
import re

import numpy as np


def get_vasp_version(string):
    """Extract version number from header of stdout.

    Example::

      >>> get_vasp_version('potato vasp.6.1.2 bumblebee')
      '6.1.2'

    """
    match = re.search(r'vasp\.(\S+)', string, re.M)
    return match.group(1)


class VaspChargeDensity:
    """Class for representing VASP charge density.

    Filename is normally CHG."""
    # Can the filename be CHGCAR?  There's a povray tutorial
    # in doc/tutorials where it's CHGCAR as of January 2021.  --askhl

    def __init__(self, filename):
        # Instance variables
        self.atoms = []  # List of Atoms objects
        self.chg = []  # Charge density
        self.chgdiff = []  # Charge density difference, if spin polarized
        self.aug = ''  # Augmentation charges, not parsed just a big string
        self.augdiff = ''  # Augmentation charge differece, is spin polarized

        # Note that the augmentation charge is not a list, since they
        # are needed only for CHGCAR files which store only a single
        # image.
        if filename is not None:
            self.read(filename)

    def is_spin_polarized(self):
        if len(self.chgdiff) > 0:
            return True
        return False

    def _read_chg(self, fobj, chg, volume):
        """Read charge from file object

        Utility method for reading the actual charge density (or
        charge density difference) from a file object. On input, the
        file object must be at the beginning of the charge block, on
        output the file position will be left at the end of the
        block. The chg array must be of the correct dimensions.

        """
        # VASP writes charge density as
        # WRITE(IU,FORM) (((C(NX,NY,NZ),NX=1,NGXC),NY=1,NGYZ),NZ=1,NGZC)
        # Fortran nested implied do loops; innermost index fastest
        # First, just read it in
        for zz in range(chg.shape[2]):
            for yy in range(chg.shape[1]):
                chg[:, yy, zz] = np.fromfile(fobj, count=chg.shape[0], sep=' ')
        chg /= volume

    def read(self, filename):
        """Read CHG or CHGCAR file.

        If CHG contains charge density from multiple steps all the
        steps are read and stored in the object. By default VASP
        writes out the charge density every 10 steps.

        chgdiff is the difference between the spin up charge density
        and the spin down charge density and is thus only read for a
        spin-polarized calculation.

        aug is the PAW augmentation charges found in CHGCAR. These are
        not parsed, they are just stored as a string so that they can
        be written again to a CHGCAR format file.

        """
        import ase.io.vasp as aiv
        with open(filename) as fd:
            self.atoms = []
            self.chg = []
            self.chgdiff = []
            self.aug = ''
            self.augdiff = ''
            while True:
                try:
                    atoms = aiv.read_vasp(fd)
                except (KeyError, RuntimeError, ValueError):
                    # Probably an empty line, or we tried to read the
                    # augmentation occupancies in CHGCAR
                    break
                fd.readline()
                ngr = fd.readline().split()
                ng = (int(ngr[0]), int(ngr[1]), int(ngr[2]))
                chg = np.empty(ng)
                self._read_chg(fd, chg, atoms.get_volume())
                self.chg.append(chg)
                self.atoms.append(atoms)
                # Check if the file has a spin-polarized charge density part,
                # and if so, read it in.
                fl = fd.tell()
                # First check if the file has an augmentation charge part
                # (CHGCAR file.)
                line1 = fd.readline()
                if line1 == '':
                    break
                elif line1.find('augmentation') != -1:
                    augs = [line1]
                    while True:
                        line2 = fd.readline()
                        if line2.split() == ngr:
                            self.aug = ''.join(augs)
                            augs = []
                            chgdiff = np.empty(ng)
                            self._read_chg(fd, chgdiff, atoms.get_volume())
                            self.chgdiff.append(chgdiff)
                        elif line2 == '':
                            break
                        else:
                            augs.append(line2)
                    if len(self.aug) == 0:
                        self.aug = ''.join(augs)
                        augs = []
                    else:
                        self.augdiff = ''.join(augs)
                        augs = []
                elif line1.split() == ngr:
                    chgdiff = np.empty(ng)
                    self._read_chg(fd, chgdiff, atoms.get_volume())
                    self.chgdiff.append(chgdiff)
                else:
                    fd.seek(fl)

    def _write_chg(self, fobj, chg, volume, format='chg'):
        """Write charge density

        Utility function similar to _read_chg but for writing.

        """
        # Make a 1D copy of chg, must take transpose to get ordering right
        chgtmp = chg.T.ravel()
        # Multiply by volume
        chgtmp = chgtmp * volume
        # Must be a tuple to pass to string conversion
        chgtmp = tuple(chgtmp)
        # CHG format - 10 columns
        if format.lower() == 'chg':
            # Write all but the last row
            for ii in range((len(chgtmp) - 1) // 10):
                fobj.write(' %#11.5G %#11.5G %#11.5G %#11.5G %#11.5G\
 %#11.5G %#11.5G %#11.5G %#11.5G %#11.5G\n' % chgtmp[ii * 10:(ii + 1) * 10])
            # If the last row contains 10 values then write them without a
            # newline
            if len(chgtmp) % 10 == 0:
                fobj.write(' %#11.5G %#11.5G %#11.5G %#11.5G %#11.5G'
                           ' %#11.5G %#11.5G %#11.5G %#11.5G %#11.5G' %
                           chgtmp[len(chgtmp) - 10:len(chgtmp)])
            # Otherwise write fewer columns without a newline
            else:
                for ii in range(len(chgtmp) % 10):
                    fobj.write((' %#11.5G') %
                               chgtmp[len(chgtmp) - len(chgtmp) % 10 + ii])
        # Other formats - 5 columns
        else:
            # Write all but the last row
            for ii in range((len(chgtmp) - 1) // 5):
                fobj.write(' %17.10E %17.10E %17.10E %17.10E %17.10E\n' %
                           chgtmp[ii * 5:(ii + 1) * 5])
            # If the last row contains 5 values then write them without a
            # newline
            if len(chgtmp) % 5 == 0:
                fobj.write(' %17.10E %17.10E %17.10E %17.10E %17.10E' %
                           chgtmp[len(chgtmp) - 5:len(chgtmp)])
            # Otherwise write fewer columns without a newline
            else:
                for ii in range(len(chgtmp) % 5):
                    fobj.write((' %17.10E') %
                               chgtmp[len(chgtmp) - len(chgtmp) % 5 + ii])
        # Write a newline whatever format it is
        fobj.write('\n')

    def write(self, filename, format=None):
        """Write VASP charge density in CHG format.

        filename: str
            Name of file to write to.
        format: str
            String specifying whether to write in CHGCAR or CHG
            format.

        """
        import ase.io.vasp as aiv
        if format is None:
            if filename.lower().find('chgcar') != -1:
                format = 'chgcar'
            elif filename.lower().find('chg') != -1:
                format = 'chg'
            elif len(self.chg) == 1:
                format = 'chgcar'
            else:
                format = 'chg'
        with open(filename, 'w') as fd:
            for ii, chg in enumerate(self.chg):
                if format == 'chgcar' and ii != len(self.chg) - 1:
                    continue  # Write only the last image for CHGCAR
                aiv.write_vasp(fd,
                               self.atoms[ii],
                               direct=True)
                fd.write('\n')
                for dim in chg.shape:
                    fd.write(' %4i' % dim)
                fd.write('\n')
                vol = self.atoms[ii].get_volume()
                self._write_chg(fd, chg, vol, format)
                if format == 'chgcar':
                    fd.write(self.aug)
                if self.is_spin_polarized():
                    if format == 'chg':
                        fd.write('\n')
                    for dim in chg.shape:
                        fd.write(' %4i' % dim)
                    fd.write('\n')  # a new line after dim is required
                    self._write_chg(fd, self.chgdiff[ii], vol, format)
                    if format == 'chgcar':
                        # a new line is always provided self._write_chg
                        fd.write(self.augdiff)
                if format == 'chg' and len(self.chg) > 1:
                    fd.write('\n')


class VaspDos:
    """Class for representing density-of-states produced by VASP

    The energies are in property self.energy

    Site-projected DOS is accesible via the self.site_dos method.

    Total and integrated DOS is accessible as numpy.ndarray's in the
    properties self.dos and self.integrated_dos. If the calculation is
    spin polarized, the arrays will be of shape (2, NDOS), else (1,
    NDOS).

    The self.efermi property contains the currently set Fermi
    level. Changing this value shifts the energies.

    """

    def __init__(self, doscar='DOSCAR', efermi=0.0):
        """Initialize"""
        self._efermi = 0.0
        self.read_doscar(doscar)
        self.efermi = efermi

        # we have determine the resort to correctly map ase atom index to the
        # POSCAR.
        self.sort = []
        self.resort = []
        if os.path.isfile('ase-sort.dat'):
            with open('ase-sort.dat') as file:
                lines = file.readlines()
            for line in lines:
                data = line.split()
                self.sort.append(int(data[0]))
                self.resort.append(int(data[1]))

    def _set_efermi(self, efermi):
        """Set the Fermi level."""
        ef = efermi - self._efermi
        self._efermi = efermi
        self._total_dos[0, :] = self._total_dos[0, :] - ef
        try:
            self._site_dos[:, 0, :] = self._site_dos[:, 0, :] - ef
        except IndexError:
            pass

    def _get_efermi(self):
        return self._efermi

    efermi = property(_get_efermi, _set_efermi, None, "Fermi energy.")

    def _get_energy(self):
        """Return the array with the energies."""
        return self._total_dos[0, :]

    energy = property(_get_energy, None, None, "Array of energies")

    def site_dos(self, atom, orbital):
        """Return an NDOSx1 array with dos for the chosen atom and orbital.

        atom: int
            Atom index
        orbital: int or str
            Which orbital to plot

        If the orbital is given as an integer:
        If spin-unpolarized calculation, no phase factors:
        s = 0, p = 1, d = 2
        Spin-polarized, no phase factors:
        s-up = 0, s-down = 1, p-up = 2, p-down = 3, d-up = 4, d-down = 5
        If phase factors have been calculated, orbitals are
        s, py, pz, px, dxy, dyz, dz2, dxz, dx2
        double in the above fashion if spin polarized.

        """
        # Correct atom index for resorting if we need to. This happens when the
        # ase-sort.dat file exists, and self.resort is not empty.
        if self.resort:
            atom = self.resort[atom]

        # Integer indexing for orbitals starts from 1 in the _site_dos array
        # since the 0th column contains the energies
        if isinstance(orbital, int):
            return self._site_dos[atom, orbital + 1, :]
        n = self._site_dos.shape[1]

        from .vasp_data import PDOS_orbital_names_and_DOSCAR_column
        norb = PDOS_orbital_names_and_DOSCAR_column[n]

        return self._site_dos[atom, norb[orbital.lower()], :]

    def _get_dos(self):
        if self._total_dos.shape[0] == 3:
            return self._total_dos[1, :]
        elif self._total_dos.shape[0] == 5:
            return self._total_dos[1:3, :]

    dos = property(_get_dos, None, None, 'Average DOS in cell')

    def _get_integrated_dos(self):
        if self._total_dos.shape[0] == 3:
            return self._total_dos[2, :]
        elif self._total_dos.shape[0] == 5:
            return self._total_dos[3:5, :]

    integrated_dos = property(_get_integrated_dos, None, None,
                              'Integrated average DOS in cell')

    def read_doscar(self, fname="DOSCAR"):
        """Read a VASP DOSCAR file"""
        with open(fname) as fd:
            natoms = int(fd.readline().split()[0])
            [fd.readline() for _ in range(4)]
            # First we have a block with total and total integrated DOS
            ndos = int(fd.readline().split()[2])
            dos = []
            for _ in range(ndos):
                dos.append(np.array([float(x) for x in fd.readline().split()]))
            self._total_dos = np.array(dos).T
            # Next we have one block per atom, if INCAR contains the stuff
            # necessary for generating site-projected DOS
            dos = []
            for _ in range(natoms):
                line = fd.readline()
                if line == '':
                    # No site-projected DOS
                    break
                ndos = int(line.split()[2])
                line = fd.readline().split()
                cdos = np.empty((ndos, len(line)))
                cdos[0] = np.array(line)
                for nd in range(1, ndos):
                    line = fd.readline().split()
                    cdos[nd] = np.array([float(x) for x in line])
                dos.append(cdos.T)
            self._site_dos = np.array(dos)



import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.abc import GetOutputsMixin
from ase.calculators.singlepoint import (
    SinglePointDFTCalculator,
    arrays_to_kpoints,
)
from ase.outputs import Properties, all_outputs


@pytest.fixture()
def rng():
    return np.random.RandomState(17)


@pytest.fixture()
def props(rng):
    nspins, nkpts, nbands = 2, 3, 5
    natoms = 4

    results = dict(
        natoms=natoms,
        energy=rng.random(),
        free_energy=rng.random(),
        energies=rng.random(natoms),
        forces=rng.random((natoms, 3)),
        stress=rng.random(6),
        stresses=rng.random((natoms, 6)),
        nspins=nspins,
        nkpts=nkpts,
        nbands=nbands,
        eigenvalues=rng.random((nspins, nkpts, nbands)),
        occupations=rng.random((nspins, nkpts, nbands)),
        fermi_level=rng.random(),
        ibz_kpoints=rng.random((nkpts, 3)),
        kpoint_weights=rng.random(nkpts),
        dipole=rng.random(3),
        charges=rng.random(natoms),
        magmom=rng.random(),
        magmoms=rng.random(natoms),
        polarization=rng.random(3),
        dielectric_tensor=rng.random((3, 3)),
        born_effective_charges=rng.random((natoms, 3, 3)),
    )
    return Properties(results)


def test_properties_big(props):
    for name in all_outputs:
        assert name in props, name
        obj = props[name]
        print(name, obj)


def test_singlepoint_roundtrip(props):

    atoms = bulk('Au') * (1, 1, props['natoms'])

    kpts = arrays_to_kpoints(props['eigenvalues'], props['occupations'],
                             props['kpoint_weights'])
    calc = SinglePointDFTCalculator(atoms=atoms, kpts=kpts,
                                    efermi=props['fermi_level'],
                                    forces=props['forces'])

    props1 = calc.properties()
    print(props1)

    assert set(props1) >= {
        'eigenvalues', 'occupations', 'kpoint_weights', 'fermi_level'}

    for prop in props1:
        assert props[prop] == pytest.approx(props1[prop])


def test_output_mixin(props, rng):

    class OutputsMixinTester(GetOutputsMixin):
        def __init__(self, props):
            self.results = props

        def _outputmixin_get_results(self):
            return self.results

    tester = OutputsMixinTester(props)

    for getter, key in [('get_fermi_level', 'fermi_level'),
                        ('get_ibz_k_points', 'ibz_kpoints'),
                        ('get_k_point_weights', 'kpoint_weights'),
                        ('get_number_of_bands', 'nbands'),
                        ('get_number_of_spins', 'nspins')]:

        assert getattr(tester, getter)() == pytest.approx(props[key])

        assert tester.get_spin_polarized() is True

    for spin in range(props['nspins']):
        for kpt_index in rng.choice(range(props['nkpts']), size=2):
            assert (tester.get_eigenvalues(kpt=kpt_index, spin=spin)
                    ==
                    pytest.approx(props['eigenvalues'][spin][kpt_index]))
            assert (tester.get_occupation_numbers(kpt=kpt_index, spin=spin)
                    ==
                    pytest.approx(props['occupations'][spin][kpt_index]))



import numpy as np
import pytest

import ase
import ase.lattice.hexagonal
from ase.build import bulk, molecule
from ase.neighborlist import (
    first_neighbors,
    mic,
    neighbor_list,
    primitive_neighbor_list,
)


@pytest.mark.slow()
def test_neighbor_kernel():
    tol = 1e-7

    # two atoms
    a = ase.Atoms('CC', positions=[[0.5, 0.5, 0.5], [1, 1, 1]],
                  cell=[10, 10, 10],
                  pbc=True)
    i, j, d = neighbor_list("ijd", a, 1.1)
    assert (i == np.array([0, 1])).all()
    assert (j == np.array([1, 0])).all()
    assert np.abs(d - np.array([np.sqrt(3 / 4), np.sqrt(3 / 4)])).max() < tol

    # test_neighbor_list
    for pbc in [True, False, [True, False, True]]:
        a = ase.Atoms('4001C', cell=[29, 29, 29])
        a.set_scaled_positions(np.transpose([np.random.random(len(a)),
                                             np.random.random(len(a)),
                                             np.random.random(len(a))]))
        j, dr, i, abs_dr, shift = neighbor_list("jDidS", a, 1.85)

        assert (np.bincount(i) == np.bincount(j)).all()

        r = a.get_positions()
        dr_direct = mic(r[j] - r[i], a.cell)
        assert np.abs(r[j] - r[i] + shift.dot(a.cell) - dr_direct).max() < tol

        abs_dr_from_dr = np.sqrt(np.sum(dr * dr, axis=1))
        abs_dr_direct = np.sqrt(np.sum(dr_direct * dr_direct, axis=1))

        assert np.all(np.abs(abs_dr - abs_dr_from_dr) < 1e-12)
        assert np.all(np.abs(abs_dr - abs_dr_direct) < 1e-12)

        assert np.all(np.abs(dr - dr_direct) < 1e-12)

    # test_neighbor_list_atoms_outside_box
    for pbc in [True, False, [True, False, True]]:
        a = ase.Atoms('4001C', cell=[29, 29, 29])
        a.set_scaled_positions(np.transpose([np.random.random(len(a)),
                                             np.random.random(len(a)),
                                             np.random.random(len(a))]))
        a.set_pbc(pbc)
        a.positions[100, :] += a.cell[0, :]
        a.positions[200, :] += a.cell[1, :]
        a.positions[300, :] += a.cell[2, :]
        j, dr, i, abs_dr, shift = neighbor_list("jDidS", a, 1.85)

        assert (np.bincount(i) == np.bincount(j)).all()

        r = a.get_positions()
        dr_direct = mic(r[j] - r[i], a.cell)
        assert np.abs(r[j] - r[i] + shift.dot(a.cell) - dr_direct).max() < tol

        abs_dr_from_dr = np.sqrt(np.sum(dr * dr, axis=1))
        abs_dr_direct = np.sqrt(np.sum(dr_direct * dr_direct, axis=1))

        assert np.all(np.abs(abs_dr - abs_dr_from_dr) < 1e-12)
        assert np.all(np.abs(abs_dr - abs_dr_direct) < 1e-12)

        assert np.all(np.abs(dr - dr_direct) < 1e-12)

    # test_small_cell
    a = ase.Atoms('C', positions=[[0.5, 0.5, 0.5]], cell=[1, 1, 1],
                  pbc=True)
    i, j, dr, shift = neighbor_list("ijDS", a, 1.1)
    assert np.bincount(i)[0] == 6
    assert (dr == shift).all()

    i, j = neighbor_list("ij", a, 1.5)
    assert np.bincount(i)[0] == 18

    a.set_pbc(False)
    i = neighbor_list("i", a, 1.1)
    assert len(i) == 0

    a.set_pbc([True, False, False])
    i = neighbor_list("i", a, 1.1)
    assert np.bincount(i)[0] == 2

    a.set_pbc([True, False, True])
    i = neighbor_list("i", a, 1.1)
    assert np.bincount(i)[0] == 4

    # test_out_of_cell_small_cell
    a = ase.Atoms('CC', positions=[[0.5, 0.5, 0.5],
                                   [1.1, 0.5, 0.5]],
                  cell=[1, 1, 1], pbc=False)
    i1, j1, r1 = neighbor_list("ijd", a, 1.1)
    a.set_cell([2, 1, 1])
    i2, j2, r2 = neighbor_list("ijd", a, 1.1)

    assert (i1 == i2).all()
    assert (j1 == j2).all()
    assert np.abs(r1 - r2).max() < tol

    # test_out_of_cell_large_cell
    a = ase.Atoms('CC', positions=[[9.5, 0.5, 0.5],
                                   [10.1, 0.5, 0.5]],
                  cell=[10, 10, 10], pbc=False)
    i1, j1, r1 = neighbor_list("ijd", a, 1.1)
    a.set_cell([20, 10, 10])
    i2, j2, r2 = neighbor_list("ijd", a, 1.1)

    assert (i1 == i2).all()
    assert (j1 == j2).all()
    assert np.abs(r1 - r2).max() < tol

    # test_hexagonal_cell
    for sx in range(3):
        a = ase.lattice.hexagonal.Graphite('C', latticeconstant=(2.5, 10.0),
                                           size=[sx + 1, sx + 1, 1])
        i = neighbor_list("i", a, 1.85)
        assert np.all(np.bincount(i) == 3)

    # test_first_neighbors
    i = [1, 1, 1, 1, 3, 3, 3]
    assert (first_neighbors(5, i) == np.array([0, 0, 4, 4, 7, 7])).all()
    i = [0, 1, 2, 3, 4, 5]
    assert (first_neighbors(6, i) == np.array([0, 1, 2, 3, 4, 5, 6])).all()

    # test_multiple_elements
    a = molecule('HCOOH')
    a.center(vacuum=5.0)
    i = neighbor_list("i", a, 1.85)
    assert (np.bincount(i) == np.array([2, 3, 1, 1, 1])).all()

    cutoffs = {(1, 6): 1.2}
    i = neighbor_list("i", a, cutoffs)
    assert (np.bincount(i) == np.array([0, 1, 0, 0, 1])).all()

    cutoffs = {(6, 8): 1.4}
    i = neighbor_list("i", a, cutoffs)
    assert (np.bincount(i) == np.array([1, 2, 1])).all()

    cutoffs = {('H', 'C'): 1.2, (6, 8): 1.4}
    i = neighbor_list("i", a, cutoffs)
    assert (np.bincount(i) == np.array([1, 3, 1, 0, 1])).all()

    cutoffs = [0.0, 0.9, 0.0, 0.5, 0.5]
    i = neighbor_list("i", a, cutoffs)
    assert (np.bincount(i) == np.array([0, 1, 0, 0, 1])).all()

    cutoffs = [0.7, 0.9, 0.7, 0.5, 0.5]
    i = neighbor_list("i", a, cutoffs)
    assert (np.bincount(i) == np.array([2, 3, 1, 1, 1])).all()

    # test_noncubic
    a = bulk("Al", cubic=False)
    i, j, d = neighbor_list("ijd", a, 3.1)
    assert (np.bincount(i) == np.array([12])).all()
    assert np.abs(d - [2.86378246] * 12).max() < tol

    # test pbc
    nat = 10
    atoms = ase.Atoms(numbers=range(nat),
                      cell=[(0.2, 1.2, 1.4),
                            (1.4, 0.1, 1.6),
                            (1.3, 2.0, -0.1)])
    atoms.set_scaled_positions(3 * np.random.random((nat, 3)) - 1)

    for p1 in range(2):
        for p2 in range(2):
            for p3 in range(2):
                atoms.set_pbc((p1, p2, p3))
                i, j, d, D, S = neighbor_list(
                    "ijdDS", atoms, atoms.numbers * 0.2 + 0.5)
                c = np.bincount(i, minlength=len(atoms))
                atoms2 = atoms.repeat((p1 + 1, p2 + 1, p3 + 1))
                i2, j2, d2, _D2, _S2 = neighbor_list(
                    "ijdDS", atoms2, atoms2.numbers * 0.2 + 0.5)
                c2 = np.bincount(i2, minlength=len(atoms))
                c2.shape = (-1, nat)
                dd = d.sum() * (p1 + 1) * (p2 + 1) * (p3 + 1) - d2.sum()
                pos = atoms.positions
                dr = np.linalg.solve(
                    atoms.cell.T, (pos[1] - pos[0]).T).T + np.array([0, 0, 3])
                assert abs(dd) < 1e-10
                assert not (c2 - c).any()

    c = 0.0058
    i, j, d = primitive_neighbor_list('ijd',
                                      [True, True, True],
                                      np.eye(3) * 7.56,
                                      np.array([[0, 0, 0],
                                                [0, 0, 0.99875]]),
                                      [c, c],
                                      self_interaction=False,
                                      use_scaled_positions=True)
    assert np.all(i == [0, 1])
    assert np.all(j == [1, 0])
    assert np.allclose(d, [0.00945, 0.00945])

    # Empty atoms object
    i, D, d, j, S = neighbor_list("iDdjS", ase.Atoms(), 1.0)
    assert i.dtype == int
    assert j.dtype == int
    assert d.dtype == float
    assert D.dtype == float
    assert S.dtype == int
    assert i.shape == (0,)
    assert j.shape == (0,)
    assert d.shape == (0,)
    assert D.shape == (0, 3)
    assert S.shape == (0, 3)

    # Check that only a scalar (not a tuple) is returned if we request a single
    # argument.
    i = neighbor_list("i", ase.Atoms(), 1.0)
    assert i.dtype == int
    assert i.shape == (0,)



import numpy as np

from ase import Atoms
from ase.neighborlist import natural_cutoffs


def test_cutoffs():
    atoms = Atoms("HCOPtAu")

    assert np.allclose(natural_cutoffs(atoms),
                       [0.31, 0.76, 0.66, 1.36, 1.36])
    assert np.allclose(natural_cutoffs(atoms, mult=1.2),
                       [0.372, 0.912, 0.792, 1.632, 1.632])
    assert np.allclose(natural_cutoffs(atoms, mult=1.2, Au=1),
                       [0.372, 0.912, 0.792, 1.632, 1])



# type: ignore
"""
This module defines an ASE interface to Turbomole: http://www.turbomole.com/

QMMM functionality provided by Markus Kaukonen <markus.kaukonen@iki.fi>.

Please read the license file (../../LICENSE)

Contact: Ivan Kondov <ivan.kondov@kit.edu>
"""
import os
import re
import warnings
from math import floor, log10

import numpy as np

from ase.calculators.calculator import (
    Calculator,
    PropertyNotImplementedError,
    ReadError,
)
from ase.calculators.turbomole import reader
from ase.calculators.turbomole.executor import execute, get_output_filename
from ase.calculators.turbomole.parameters import TurbomoleParameters
from ase.calculators.turbomole.reader import read_convergence, read_data_group
from ase.calculators.turbomole.writer import add_data_group, delete_data_group
from ase.io import read, write
from ase.units import Bohr, Ha


class TurbomoleOptimizer:
    def __init__(self, atoms, calc):
        self.atoms = atoms
        self.calc = calc
        self.atoms.calc = self.calc

    def todict(self):
        return {'type': 'optimization',
                'optimizer': 'TurbomoleOptimizer'}

    def run(self, fmax=None, steps=None):
        if fmax is not None:
            self.calc.parameters['force convergence'] = fmax
            self.calc.parameters.verify()
        if steps is not None:
            self.calc.parameters['geometry optimization iterations'] = steps
            self.calc.parameters.verify()
        self.calc.calculate()
        self.atoms.positions[:] = self.calc.atoms.positions
        self.calc.parameters['task'] = 'energy'


class Turbomole(Calculator):

    """constants"""
    name = 'Turbomole'

    implemented_properties = ['energy', 'forces', 'dipole', 'free_energy',
                              'charges']

    tm_files = [
        'control', 'coord', 'basis', 'auxbasis', 'energy', 'gradient', 'mos',
        'alpha', 'beta', 'statistics', 'GEO_OPT_CONVERGED', 'GEO_OPT_FAILED',
        'not.converged', 'nextstep', 'hessapprox', 'job.last', 'job.start',
        'optinfo', 'statistics', 'converged', 'vibspectrum',
        'vib_normal_modes', 'hessian', 'dipgrad', 'dscf_problem', 'pc.txt',
        'pc_gradients.txt'
    ]
    tm_tmp_files = [
        'errvec', 'fock', 'oldfock', 'dens', 'ddens', 'diff_densmat',
        'diff_dft_density', 'diff_dft_oper', 'diff_fockmat', 'diis_errvec',
        'diis_oldfock', 'dh'
    ]

    # initialize attributes
    results = {}
    initialized = False
    pc_initialized = False
    converged = False
    updated = False
    update_energy = None
    update_forces = None
    update_geometry = None
    update_hessian = None
    atoms = None
    forces = None
    e_total = None
    dipole = None
    charges = None
    version = None
    runtime = None
    datetime = None
    hostname = None
    pcpot = None

    def __init__(self, label=None, calculate_energy='dscf',
                 calculate_forces='grad', post_HF=False, atoms=None,
                 restart=False, define_str=None, control_kdg=None,
                 control_input=None, reset_tolerance=1e-2, **kwargs):

        super().__init__(label=label)
        self.parameters = TurbomoleParameters()

        self.calculate_energy = calculate_energy
        self.calculate_forces = calculate_forces
        self.post_HF = post_HF
        self.restart = restart
        self.parameters.define_str = define_str
        self.control_kdg = control_kdg
        self.control_input = control_input
        self.reset_tolerance = reset_tolerance

        if self.restart:
            self._set_restart(kwargs)
        else:
            self.parameters.update(kwargs)
            self.set_parameters()
            self.parameters.verify()
            self.reset()

        if atoms is not None:
            atoms.calc = self
            self.set_atoms(atoms)

    def __getitem__(self, item):
        return getattr(self, item)

    def _set_restart(self, params_update):
        """constructs atoms, parameters and results from a previous
        calculation"""

        # read results, key parameters and non-key parameters
        self.read_restart()
        self.parameters.read_restart(self.atoms, self.results)

        # update and verify parameters
        self.parameters.update_restart(params_update)
        self.set_parameters()
        self.parameters.verify()

        # if a define string is specified by user then run define
        if self.parameters.define_str:
            execute(['define'], input_str=self.parameters.define_str)

        self._set_post_define()

        self.initialized = True
        # more precise convergence tests are necessary to set these flags:
        self.update_energy = True
        self.update_forces = True
        self.update_geometry = True
        self.update_hessian = True

    def _set_post_define(self):
        """non-define keys, user-specified changes in the control file"""
        self.parameters.update_no_define_parameters()

        # delete user-specified data groups
        if self.control_kdg:
            for dg in self.control_kdg:
                delete_data_group(dg)

        # append user-defined input to control
        if self.control_input:
            for inp in self.control_input:
                add_data_group(inp, raw=True)

        # add point charges if pcpot defined:
        if self.pcpot:
            self.set_point_charges()

    def set_parameters(self):
        """loads the default parameters and updates with actual values"""
        if self.parameters.get('use resolution of identity'):
            self.calculate_energy = 'ridft'
            self.calculate_forces = 'rdgrad'

    def reset(self):
        """removes all turbomole input, output and scratch files,
        and deletes results dict and the atoms object"""
        self.atoms = None
        self.results = {}
        self.results['calculation parameters'] = {}
        ase_files = [
            f for f in os.listdir(
                self.directory) if f.startswith('ASE.TM.')]
        for f in self.tm_files + self.tm_tmp_files + ase_files:
            if os.path.exists(f):
                os.remove(f)
        self.initialized = False
        self.pc_initialized = False
        self.converged = False

    def set_atoms(self, atoms):
        """Create the self.atoms object and writes the coord file. If
        self.atoms exists a check for changes and an update of the atoms
        is performed. Note: Only positions changes are tracked in this
        version.
        """
        changes = self.check_state(atoms, tol=1e-13)
        if self.atoms == atoms or 'positions' not in changes:
            # print('two atoms obj are (almost) equal')
            if self.updated and os.path.isfile('coord'):
                self.updated = False
                a = read('coord').get_positions()
                if np.allclose(a, atoms.get_positions(), rtol=0, atol=1e-13):
                    return
            else:
                return

        changes = self.check_state(atoms, tol=self.reset_tolerance)
        if 'positions' in changes:
            # print(two atoms obj are different')
            self.reset()
        else:
            # print('two atoms obj are slightly different')
            if self.parameters['use redundant internals']:
                self.reset()

        write('coord', atoms)
        self.atoms = atoms.copy()
        self.update_energy = True
        self.update_forces = True
        self.update_geometry = True
        self.update_hessian = True

    def initialize(self):
        """prepare turbomole control file by running module 'define'"""
        if self.initialized:
            return
        self.parameters.verify()
        if not self.atoms:
            raise RuntimeError('atoms missing during initialization')
        if not os.path.isfile('coord'):
            raise OSError('file coord not found')

        # run define
        define_str = self.parameters.get_define_str(len(self.atoms))
        execute(['define'], input_str=define_str)

        # process non-default initial guess
        iguess = self.parameters['initial guess']
        if isinstance(iguess, dict) and 'use' in iguess.keys():
            # "use" initial guess
            if self.parameters['multiplicity'] != 1 or self.parameters['uhf']:
                define_str = '\n\n\ny\nuse ' + iguess['use'] + '\nn\nn\nq\n'
            else:
                define_str = '\n\n\ny\nuse ' + iguess['use'] + '\nn\nq\n'
            execute(['define'], input_str=define_str)
        elif self.parameters['initial guess'] == 'hcore':
            # "hcore" initial guess
            if self.parameters['multiplicity'] != 1 or self.parameters['uhf']:
                delete_data_group('uhfmo_alpha')
                delete_data_group('uhfmo_beta')
                add_data_group('uhfmo_alpha', 'none file=alpha')
                add_data_group('uhfmo_beta', 'none file=beta')
            else:
                delete_data_group('scfmo')
                add_data_group('scfmo', 'none file=mos')

        self._set_post_define()

        self.initialized = True
        self.converged = False

    def calculation_required(self, atoms, properties):
        if self.atoms != atoms:
            return True
        for prop in properties:
            if prop == 'energy' and self.e_total is None:
                return True
            elif prop == 'forces' and self.forces is None:
                return True
        return False

    def calculate(self, atoms=None):
        """execute the requested job"""
        if atoms is None:
            atoms = self.atoms
        if self.parameters['task'] in ['energy', 'energy calculation']:
            self.get_potential_energy(atoms)
        if self.parameters['task'] in ['gradient', 'gradient calculation']:
            self.get_forces(atoms)
        if self.parameters['task'] in ['optimize', 'geometry optimization']:
            self.relax_geometry(atoms)
        if self.parameters['task'] in ['frequencies', 'normal mode analysis']:
            self.normal_mode_analysis(atoms)
        self.read_results()

    def relax_geometry(self, atoms=None):
        """execute geometry optimization with script jobex"""
        if atoms is None:
            atoms = self.atoms
        self.set_atoms(atoms)
        if self.converged and not self.update_geometry:
            return
        self.initialize()
        jobex_command = ['jobex']
        if self.parameters['transition vector']:
            jobex_command.append('-trans')
        if self.parameters['use resolution of identity']:
            jobex_command.append('-ri')
        if self.parameters['force convergence']:
            par = self.parameters['force convergence']
            conv = floor(-log10(par / Ha * Bohr))
            jobex_command.extend(['-gcart', str(int(conv))])
        if self.parameters['energy convergence']:
            par = self.parameters['energy convergence']
            conv = floor(-log10(par / Ha))
            jobex_command.extend(['-energy', str(int(conv))])
        geom_iter = self.parameters['geometry optimization iterations']
        if geom_iter is not None:
            assert isinstance(geom_iter, int)
            jobex_command.extend(['-c', str(geom_iter)])
        self.converged = False
        execute(jobex_command)
        # check convergence
        self.converged = read_convergence(self.restart, self.parameters)
        if self.converged:
            self.update_energy = False
            self.update_forces = False
            self.update_geometry = False
            self.update_hessian = True
        # read results
        new_struct = read('coord')
        atoms.set_positions(new_struct.get_positions())
        self.atoms = atoms.copy()
        self.read_energy()

    def normal_mode_analysis(self, atoms=None):
        """execute normal mode analysis with modules aoforce or NumForce"""
        from ase.constraints import FixAtoms
        if atoms is None:
            atoms = self.atoms
        self.set_atoms(atoms)
        self.initialize()
        if self.update_energy:
            self.get_potential_energy(atoms)
        if self.update_hessian:
            fixatoms = []
            for constr in atoms.constraints:
                if isinstance(constr, FixAtoms):
                    ckwargs = constr.todict()['kwargs']
                    if 'indices' in ckwargs.keys():
                        fixatoms.extend(ckwargs['indices'])
            if self.parameters['numerical hessian'] is None:
                if len(fixatoms) > 0:
                    define_str = '\n\ny\n'
                    for index in fixatoms:
                        define_str += 'm ' + str(index + 1) + ' 999.99999999\n'
                    define_str += '*\n*\nn\nq\n'
                    execute(['define'], input_str=define_str)
                    dg = read_data_group('atoms')
                    regex = r'(mass\s*=\s*)999.99999999'
                    dg = re.sub(regex, r'\g<1>9999999999.9', dg)
                    dg += '\n'
                    delete_data_group('atoms')
                    add_data_group(dg, raw=True)
                execute(['aoforce'])
            else:
                nforce_cmd = ['NumForce']
                pdict = self.parameters['numerical hessian']
                if self.parameters['use resolution of identity']:
                    nforce_cmd.append('-ri')
                if len(fixatoms) > 0:
                    nforce_cmd.extend(['-frznuclei', '-central', '-c'])
                if 'central' in pdict.keys():
                    nforce_cmd.append('-central')
                if 'delta' in pdict.keys():
                    nforce_cmd.extend(['-d', str(pdict['delta'] / Bohr)])
                if self.update_forces:
                    self.get_forces(atoms)
                execute(nforce_cmd)
            self.update_hessian = False

    def read_restart(self):
        """read a previous calculation from control file"""
        self.atoms = read('coord')
        self.atoms.calc = self
        self.converged = read_convergence(self.restart, self.parameters)
        self.read_results()

    def read_results(self):
        """read all results and load them in the results entity"""
        read_methods = [
            self.read_energy,
            self.read_gradient,
            self.read_forces,
            self.read_basis_set,
            self.read_ecps,
            self.read_mos,
            self.read_occupation_numbers,
            self.read_dipole_moment,
            self.read_ssquare,
            self.read_hessian,
            self.read_vibrational_reduced_masses,
            self.read_normal_modes,
            self.read_vibrational_spectrum,
            self.read_charges,
            self.read_point_charges,
            self.read_run_parameters
        ]
        for method in read_methods:
            try:
                method()
            except ReadError as err:
                warnings.warn(err.args[0])

    def read_run_parameters(self):
        """read parameters set by define and not in self.parameters"""
        reader.read_run_parameters(self.results)

    def read_energy(self):
        """Read energy from Turbomole energy file."""
        reader.read_energy(self.results, self.post_HF)
        self.e_total = self.results['total energy']

    def read_forces(self):
        """Read forces from Turbomole gradient file."""
        self.forces = reader.read_forces(self.results, len(self.atoms))

    def read_occupation_numbers(self):
        """read occupation numbers"""
        reader.read_occupation_numbers(self.results)

    def read_mos(self):
        """read the molecular orbital coefficients and orbital energies
        from files mos, alpha and beta"""

        ans = reader.read_mos(self.results)
        if ans is not None:
            self.converged = ans

    def read_basis_set(self):
        """read the basis set"""
        reader.read_basis_set(self.results)

    def read_ecps(self):
        """read the effective core potentials"""
        reader.read_ecps(self.results)

    def read_gradient(self):
        """read all information in file 'gradient'"""
        reader.read_gradient(self.results)

    def read_hessian(self):
        """Read in the hessian matrix"""
        reader.read_hessian(self.results)

    def read_normal_modes(self):
        """Read in vibrational normal modes"""
        reader.read_normal_modes(self.results)

    def read_vibrational_reduced_masses(self):
        """Read vibrational reduced masses"""
        reader.read_vibrational_reduced_masses(self.results)

    def read_vibrational_spectrum(self):
        """Read the vibrational spectrum"""
        reader.read_vibrational_spectrum(self.results)

    def read_ssquare(self):
        """Read the expectation value of S^2 operator"""
        reader.read_ssquare(self.results)

    def read_dipole_moment(self):
        """Read the dipole moment"""
        reader.read_dipole_moment(self.results)
        dip_vec = self.results['electric dipole moment']['vector']['array']
        self.dipole = np.array(dip_vec) * Bohr

    def read_charges(self):
        """read partial charges on atoms from an ESP fit"""
        filename = get_output_filename(self.calculate_energy)
        self.charges = reader.read_charges(filename, len(self.atoms))

    def get_version(self):
        """get the version of the installed turbomole package"""
        if not self.version:
            self.version = reader.read_version(self.directory)
        return self.version

    def get_datetime(self):
        """get the timestamp of most recent calculation"""
        if not self.datetime:
            self.datetime = reader.read_datetime(self.directory)
        return self.datetime

    def get_runtime(self):
        """get the total runtime of calculations"""
        if not self.runtime:
            self.runtime = reader.read_runtime(self.directory)
        return self.runtime

    def get_hostname(self):
        """get the hostname of the computer on which the calc has run"""
        if not self.hostname:
            self.hostname = reader.read_hostname(self.directory)
        return self.hostname

    def get_optimizer(self, atoms, trajectory=None, logfile=None):
        """returns a TurbomoleOptimizer object"""
        self.parameters['task'] = 'optimize'
        self.parameters.verify()
        return TurbomoleOptimizer(atoms, self)

    def get_results(self):
        """returns the results dictionary"""
        return self.results

    def get_potential_energy(self, atoms, force_consistent=True):
        # update atoms
        self.updated = self.e_total is None
        self.set_atoms(atoms)
        self.initialize()
        # if update of energy is necessary
        if self.update_energy:
            # calculate energy
            execute([self.calculate_energy])
            # check convergence
            self.converged = read_convergence(self.restart, self.parameters)
            if not self.converged:
                return None
            # read energy
            self.read_energy()

        self.update_energy = False
        return self.e_total

    def get_forces(self, atoms):
        # update atoms
        self.updated = self.forces is None
        self.set_atoms(atoms)
        # complete energy calculations
        if self.update_energy:
            self.get_potential_energy(atoms)
        # if update of forces is necessary
        if self.update_forces:
            # calculate forces
            execute([self.calculate_forces])
            # read forces
            self.read_forces()

        self.update_forces = False
        return self.forces.copy()

    def get_dipole_moment(self, atoms):
        self.get_potential_energy(atoms)
        self.read_dipole_moment()
        return self.dipole

    def get_property(self, name, atoms=None, allow_calculation=True):
        """return the value of a property"""

        if name not in self.implemented_properties:
            # an ugly work around; the caller should test the raised error
            # if name in ['magmom', 'magmoms', 'charges', 'stress']:
            # return None
            raise PropertyNotImplementedError(name)

        if atoms is None:
            atoms = self.atoms.copy()

        persist_property = {
            'energy': 'e_total',
            'forces': 'forces',
            'dipole': 'dipole',
            'free_energy': 'e_total',
            'charges': 'charges'
        }
        property_getter = {
            'energy': self.get_potential_energy,
            'forces': self.get_forces,
            'dipole': self.get_dipole_moment,
            'free_energy': self.get_potential_energy,
            'charges': self.get_charges
        }
        getter_args = {
            'energy': [atoms],
            'forces': [atoms],
            'dipole': [atoms],
            'free_energy': [atoms, True],
            'charges': [atoms]
        }

        if allow_calculation:
            result = property_getter[name](*getter_args[name])
        else:
            if hasattr(self, persist_property[name]):
                result = getattr(self, persist_property[name])
            else:
                result = None

        if isinstance(result, np.ndarray):
            result = result.copy()
        return result

    def get_charges(self, atoms):
        """return partial charges on atoms from an ESP fit"""
        self.get_potential_energy(atoms)
        self.read_charges()
        return self.charges

    def get_forces_on_point_charges(self):
        """return forces acting on point charges"""
        self.get_forces(self.atoms)
        lines = read_data_group('point_charge_gradients').split('\n')[1:]
        forces = []
        for line in lines:
            linef = line.strip().replace('D', 'E')
            forces.append([float(x) for x in linef.split()])
        # Note the '-' sign for turbomole, to get forces
        return -np.array(forces) * Ha / Bohr

    def set_point_charges(self, pcpot=None):
        """write external point charges to control"""
        if pcpot is not None and pcpot != self.pcpot:
            self.pcpot = pcpot
        if self.pcpot.mmcharges is None or self.pcpot.mmpositions is None:
            raise RuntimeError('external point charges not defined')

        if not self.pc_initialized:
            if len(read_data_group('point_charges')) == 0:
                add_data_group('point_charges', 'file=pc.txt')
            if len(read_data_group('point_charge_gradients')) == 0:
                add_data_group(
                    'point_charge_gradients',
                    'file=pc_gradients.txt'
                )
            drvopt = read_data_group('drvopt')
            if 'point charges' not in drvopt:
                drvopt += '\n   point charges\n'
                delete_data_group('drvopt')
                add_data_group(drvopt, raw=True)
            self.pc_initialized = True

        if self.pcpot.updated:
            with open('pc.txt', 'w') as pcfile:
                pcfile.write('$point_charges nocheck list\n')
                for (x, y, z), charge in zip(
                        self.pcpot.mmpositions, self.pcpot.mmcharges):
                    pcfile.write('%20.14f  %20.14f  %20.14f  %20.14f\n'
                                 % (x / Bohr, y / Bohr, z / Bohr, charge))
                pcfile.write('$end \n')
            self.pcpot.updated = False

    def read_point_charges(self):
        """read point charges from previous calculation"""
        charges, positions = reader.read_point_charges()
        if len(charges) > 0:
            self.pcpot = PointChargePotential(charges, positions)

    def embed(self, charges=None, positions=None):
        """embed atoms in an array of point-charges; function used in
            qmmm calculations."""
        self.pcpot = PointChargePotential(charges, positions)
        return self.pcpot


class PointChargePotential:
    """Point-charge potential for Turbomole"""

    def __init__(self, mmcharges, mmpositions=None):
        self.mmcharges = mmcharges
        self.mmpositions = mmpositions
        self.mmforces = None
        self.updated = True

    def set_positions(self, mmpositions):
        """set the positions of point charges"""
        self.mmpositions = mmpositions
        self.updated = True

    def set_charges(self, mmcharges):
        """set the values of point charges"""
        self.mmcharges = mmcharges
        self.updated = True

    def get_forces(self, calc):
        """forces acting on point charges"""
        self.mmforces = calc.get_forces_on_point_charges()
        return self.mmforces



import functools
import re
from collections import Counter
from dataclasses import dataclass
from fractions import Fraction
from itertools import chain, combinations, product
from typing import List, Tuple, Union

import numpy as np
from scipy.linalg import null_space

from ase.formula import Formula
from ase.units import kB

CONST = kB * np.log(10)  # Nernst constant

PREDEF_ENERGIES = {      # Default chemical potentials
    'H+': 0.0,           # for water, protons and electrons
    'e-': 0.0,
    'H2O': -2.4583
}

U_STD_AGCL = 0.222  # Standard redox potential of AgCl electrode
U_STD_SCE = 0.244   # Standard redox potential of SCE electrode


def parse_formula(formula: str, fmt: str = 'metal'):
    aq = formula.endswith('(aq)')
    charge = formula.count('+') - formula.count('-')
    formula_strip = formula.replace('(aq)', '').rstrip('+-')
    formula_obj = Formula(formula_strip, format=fmt)
    return formula_obj, charge, aq


def initialize_refs(refs_dct, reduce=False, fmt='metal'):
    """Convert dictionary entries to Species instances"""
    refs = {}
    for label, energy in refs_dct.items():
        spec = Species.from_string(label, energy, reduce, fmt)
        refs[label] = spec
    return refs


def get_product_combos(reactant, refs):
    """Obtain all possible combinations of products.

    Obtain - from the available references and based
    on the stoichiometry of the target material (reactant) -
    different combinations of products to be inserted
    in (electro)chemical reactions.
    """
    ref_elem = reactant._main_elements
    allcombos = [[] for _ in range(len(ref_elem))]
    for ref in refs.values():
        contained = ref.contains(ref_elem)
        for w in np.argwhere(contained).flatten():
            allcombos[w].append(ref)
    return [np.unique(combo) for combo in product(*allcombos)]


def get_phases(reactant, refs, T, conc, reference, tol=1e-3):
    """Obtain all the possible decomposition pathways
       for a given reactant as a collection of RedOx objects.
    """
    phases = []
    phase_matrix = []

    for products in get_product_combos(reactant, refs):
        phase = RedOx.from_species(
            reactant, products, T, conc, reference, tol
        )
        if phase is not None:
            phases.append(phase)
            phase_matrix.append(phase._vector)

    if len(phase_matrix) == 0:
        raise ValueError(
            "No valid decomposition pathways have been found" +
            " given this set of references."
        )

    return phases, np.array(phase_matrix).astype('float64')


def get_main_products(count):
    """Obtain the reaction products excluded protons,
       water and electrons.
    """
    return [spec for spec, coef in count.items()
            if coef > 0 and spec not in ['H+', 'H2O', 'e-']]


def format_label(count) -> str:
    """Obtain phase labels formatted in LaTeX math style."""
    formatted = []
    for prod in get_main_products(count):
        label = re.sub(r'(\S)([+-]+)', r'\1$^{\2}$', prod)
        label = re.sub(r'(\d+)', r'$_{\1}$', label)
        for symbol in ['+', '-']:
            count = label.count(symbol)
            if count > 1:
                label = label.replace(count * symbol, f'{count}{symbol}')
            if count == 1:
                label = label.replace(count * symbol, symbol)
        formatted.append(label)
    return ', '.join(f for f in formatted)


def make_coeff_nice(coeff, max_denom) -> str:
    """Convert a fraction into a string while limiting the denominator"""
    frac = abs(Fraction(coeff).limit_denominator(max_denom))
    if frac.numerator == frac.denominator:
        return ''
    return str(frac)


def add_numbers(ax, text) -> None:
    """Add number identifiers to the different domains of a Pourbaix diagram."""
    import matplotlib.patheffects as pfx
    for i, (x, y, _) in enumerate(text):
        txt = ax.text(
            y, x, f'{i}',
            fontsize=20,
            horizontalalignment='center'
        )
        txt.set_path_effects([pfx.withStroke(linewidth=2.0, foreground='w')])


def add_labels(ax, text) -> None:
    """Add phase labels to the different domains of a Pourbaix diagram."""
    import matplotlib.patheffects as pfx
    for i, (x, y, species) in enumerate(text):
        label = format_label(species)
        annotation = ax.annotate(
            label, xy=(y, x), color='w',
            fontsize=16, horizontalalignment='center'
        )
        annotation.set_path_effects(
            [pfx.withStroke(linewidth=2.0, foreground='k')]
        )
        annotation.draggable()
        ax.add_artist(annotation)


def wrap_text(text) -> str:
    import textwrap

    textlines = []
    for i, (_, _, species) in enumerate(text):
        label = format_label(species)
        textlines.append(
            textwrap.fill(
                f'({i})  {label}',
                width=40,
                subsequent_indent='      '
            )
        )
    return '\n'.join(textlines)


def add_phase_labels(fig, text, offset=0.0):
    """Add phase labels to the right of the diagram"""

    fig.text(
        0.75 + offset, 0.5,
        wrap_text(text),
        fontsize=16,
        va='center',
        ha='left')


def add_redox_lines(axes, pH, reference, color='k') -> None:
    """Add water redox potentials to a Pourbaix diagram"""
    const = -0.5 * PREDEF_ENERGIES['H2O']
    corr = {
        'SHE': 0,
        'RHE': 0,
        'Pt': 0,
        'AgCl': -U_STD_AGCL,
        'SCE': -U_STD_SCE,
    }
    kwargs = {
        'c': color,
        'ls': '--',
        'zorder': 2
    }
    if reference in ['SHE', 'AgCl', 'SCE']:
        slope = -59.2e-3
        axes.plot(pH, slope * pH + corr[reference], **kwargs)
        axes.plot(pH, slope * pH + const + corr[reference], **kwargs)
    elif reference in ['Pt', 'RHE']:
        axes.axhline(0 + corr[reference], **kwargs)
        axes.axhline(const + corr[reference], **kwargs)
    else:
        raise ValueError('The specified reference electrode doesnt exist')


@functools.total_ordering
class Species:
    """Class representing an individual chemical species,
       grouping relevant properties.

    Initialization
    --------------
    name: str
        A label representing the species

    formula: Formula

    charge: int
        the electric charge of the species, if ionic

    aq: bool
        whether the species is solid (False)
        or acqueous (True)

    energy: float
        the chemical potential of the species
    """
    def __init__(self,
                 name: str,
                 formula: Formula,
                 charge: int,
                 aq: bool,
                 energy: float):

        self.name = name
        self.formula = formula
        self.energy = energy
        self.charge = charge
        self.aq = aq

        self.count = formula.count()
        self.natoms = sum(self.count.values())
        self._main_elements = [
            e for e in self.count.keys() if e not in ['H', 'O']
        ]

    def __eq__(self, other):
        if not isinstance(other, Species):
            return NotImplemented
        return self.name == other.name

    def __lt__(self, other):
        if not isinstance(other, Species):
            return NotImplemented
        return self.name < other.name

    @classmethod
    def from_string(cls, string: str, energy: float,
                    reduce: bool = True, fmt: str = 'metal'):
        """Initialize the class provided a formula and an energy.

        string: str
            The chemical formula of the species (e.g. ``ZnO``).
            For solid species, the formula is automatically reduced to the
            unit formula, and the chemical potential normalized accordingly.
            Acqueous species are specified by expliciting
            all the positive or negative charges and by appending ``(aq)``.
            Parentheses for grouping functional groups are acceptable.
                e.g.
                    Be3(OH)3[3+]    ➜  wrong
                    Be3(OH)3+++     ➜  wrong
                    Be3(OH)3+++(aq) ➜  correct
                    Be3O3H3+++(aq)  ➜  correct

        energy: float
            the energy (chemical potential) associated with the species.

        reduce: bool
            reduce to the unit formula and normalize the energy accordingly.
            Formulae and energies of acqueous species are never reduced.

        fmt: str
            Formula formatting according to the available options in
            ase.formula.Formula
        """
        formula, charge, aq = parse_formula(string, fmt=fmt)
        if not aq:
            if reduce:
                formula, n_fu = formula.reduce()
                energy /= n_fu
            name = str(formula)
        else:
            name = string
        return cls(name, formula, charge, aq, energy)

    def get_chemsys(self):
        """Get the possible combinations of elements based
           on the stoichiometry. Useful for database queries.
        """
        elements = set(self.count.keys())
        elements.update(['H', 'O'])
        chemsys = list(
            chain.from_iterable(
                [combinations(elements, i + 1)
                 for i in range(len(elements))]
            )
        )
        return chemsys

    def balance_electrochemistry(self):
        """Obtain number of H2O, H+, e- "carried" by the species
           in electrochemical reactions.
        """
        n_H2O = -self.count.get('O', 0)
        n_H = -2 * n_H2O - self.count.get('H', 0)
        n_e = n_H + self.charge
        return n_H2O, n_H, n_e

    def _count_array(self, elements):
        return np.array([self.count.get(e, 0) for e in elements])

    def contains(self, elements):
        return [elem in self._main_elements for elem in elements]

    def get_fractional_composition(self, elements):
        """Obtain the fractional content of each element."""
        N_all = sum(self.count.values())
        N_elem = sum([self.count.get(e, 0) for e in elements])
        return N_elem / N_all

    def __repr__(self):
        return f'Species({self.name})'


class RedOx:
    def __init__(self, species, coeffs,
                 T=298.15, conc=1e-6,
                 reference='SHE'):
        """RedOx class representing an (electro)chemical reaction.

        Initialization
        --------------

        species: list[Species]
            The reactant and products excluded H2O, protons and electrons.

        coeffs: list[float]
            The stoichiometric coefficients of the above species.
            Positive coefficients are associated with the products,
            negative coefficients with the reactants.

        T: float
            The temperature in Kelvin. Default: 298.15 K.

        conc: float
            The concentration of ionic species. Default: 1.0e-6 M.

        reference: str
            The reference electrode. Default: SHE.
            available options: SHE, RHE, AgCl, Pt.


        Relevant methods
        ----------------

        from_species(reactant, products, T, conc, reference)
            Initialize the class from the reactant as a Species object
            and the product(s) a list of Species objects.

        equation()
            Print the chemical equation of the reaction.

        get_free_energy(U, pH):
            Obtain the reaction free energy at a given
            applied potential (U) and pH.

        """
        self.species = species
        self.coeffs = coeffs
        self.T = T
        self.conc = conc
        self.reference = reference
        self.count = Counter()

        alpha = CONST * T   # 0.059 eV @ T=298.15K
        const_term = 0
        pH_term = 0
        U_term = 0

        for spec, coef in zip(species, coeffs):
            self.count[spec.name] = coef
            amounts = spec.balance_electrochemistry()

            const_term += coef * (
                spec.energy + alpha * (spec.aq * np.log10(conc))
            )
            pH_term += - coef * alpha * amounts[1]
            U_term += - coef * amounts[2]

            for name, n in zip(['H2O', 'H+', 'e-'], amounts):
                const_term += coef * n * PREDEF_ENERGIES[name]
                self.count[name] += coef * n

        const_corr, pH_corr = self.get_ref_correction(reference, alpha)
        self._vector = [
            float(const_term + const_corr),
            float(U_term),
            float(pH_term + pH_corr)
        ]

    @classmethod
    def from_species(cls, reactant, products,
                     T: float = 298.15, conc: float = 1e-6,
                     reference: str = 'SHE', tol: float = 1e-3):
        """Initialize the class from a combination of
           reactant and products. The stoichiometric
           coefficients are automatically determined.

           reactant: Species
           products: list[Species]
        """
        reac_elem = reactant._main_elements
        reac_count = [-reactant._count_array(reac_elem)]
        prod_count = [p._count_array(reac_elem) for p in products]
        elem_matrix = np.array(reac_count + prod_count).T
        coeffs = null_space(elem_matrix).flatten()

        if len(coeffs) > 0 and all(coeffs > tol):
            coeffs /= coeffs[0]
            coeffs[0] = -coeffs[0]
            species = (reactant, *products)
            phase = cls(species, coeffs, T, conc, reference)
            return phase

        return None

    def get_ref_correction(self, reference: str,
                           alpha: float) -> Tuple[float, float]:
        """Correct the constant and pH contributions to the reaction free energy
           based on the reference electrode of choice and the temperature
           (alpha=k_B*T*ln(10))
        """
        n_e = self.count['e-']
        gibbs_corr = 0.0
        pH_corr = 0.0
        if reference in ['RHE', 'Pt']:
            pH_corr += n_e * alpha
            if reference == 'Pt' and n_e < 0:
                gibbs_corr += n_e * 0.5 * PREDEF_ENERGIES['H2O']
        if reference == 'AgCl':
            gibbs_corr -= n_e * U_STD_AGCL
        if reference == 'SCE':
            gibbs_corr -= n_e * U_STD_SCE

        return gibbs_corr, pH_corr

    def equation(self, max_denom: int = 50) -> str:
        """Print the chemical reaction."""

        reactants = []
        products = []
        for s, n in self.count.items():
            if abs(n) <= 1e-6:
                continue
            nice_coeff = make_coeff_nice(n, max_denom)
            substr = f'{nice_coeff}{s}'
            if n > 0:
                products.append(substr)
            else:
                reactants.append(substr)

        return "  ➜  ".join([" + ".join(reactants), " + ".join(products)])

    def get_free_energy(self, U: float, pH: float) -> float:
        """Evaluate the reaction free energy
           at a given applied potential U and pH"""
        return self._vector[0] + self._vector[1] * U + self._vector[2] * pH


class Pourbaix:
    """Pourbaix class for acqueous stability evaluations.

    Allows to determine the most stable phase in a given set
    of pH and potential conditions and to evaluate a complete diagram.

    Initialization
    --------------

    material_name: str
        The formula of the target material. It is preferrable
        to provide the reduced formula (e.g. RuO2 instad of Ru2O4).

    refs_dct: dict
        A dictionary containing the formulae of the target material
        and its competing phases (solid and/or ionic) as keys,
        and their formation energies as values.

    T: float
        Temperature in Kelvin. Default: 298.15 K.

    conc: float
        Concentration of the ionic species. Default: 1e-6 mol/L.

    reference: str
        The reference electrode. Default: SHE.
        available options: SHE, RHE, AgCl, Pt.


    Relevant methods
    ----------------

    get_pourbaix_energy(U, pH)
        obtain the energy of the target material
        relative to the most stable phase at a given potential U and pH.
        If negative, the target material can be regarded as stable.
    plot(...)
        plot a complete Pourbaix diagram in a given pH and potential window.


    Relevant attributes
    -------------------

    material: Species
        the target material as a Species object

    phases: list[RedOx]
        the available decomposition reactions of the target material
        into its competing phases as a list of RedOx objects.

    """
    def __init__(self,
                 material_name: str,
                 refs_dct: dict,
                 T: float = 298.15,
                 conc: float = 1.0e-6,
                 reference: str = 'SHE'):

        self.material_name = material_name
        self.refs = refs_dct
        self.T = T
        self.conc = conc
        self.reference = reference

        refs = initialize_refs(refs_dct)
        self.material = refs.pop(material_name)
        self.phases, phase_matrix = get_phases(
            self.material, refs, T, conc, reference
        )
        self._const = phase_matrix[:, 0]
        self._var = phase_matrix[:, 1:]

    def _decompose(self, U, pH):
        """Evaluate the reaction energy for decomposing
           the target material into each of the available products
           at a given pH and applied potential.
        """
        return self._const + np.dot(self._var, [U, pH])

    def _get_pourbaix_energy(self, U, pH):
        """Evaluate the Pourbaix energy"""
        energies = self._decompose(U, pH)
        i_min = np.argmin(energies)
        return -energies[i_min], i_min

    def get_pourbaix_energy(self, U, pH, verbose=False):
        """Evaluate the Pourbaix energy, print info about
        the most stable phase and the corresponding energy at
        a given potential U and pH.

        The Pourbaix energy represents the energy of the target material
        relative to the most stable competing phase. If negative,
        the target material can be considered as stable.
        """
        energy, index = self._get_pourbaix_energy(U, pH)
        phase = self.phases[index]
        if verbose:
            if energy <= 0.0:
                print(f'{self.material.name} is stable.')
            else:
                print(f'Stable phase: \n{phase.equation()}')
            print(f'Energy: {energy:.3f} eV')
        return energy, phase

    def get_equations(self, contains: Union[str, None] = None):
        """Print the chemical reactions of the available phases.

        the argument `contains' allows to filter for phases containing a
        particular species
            e.g. get_equations(contains='ZnO')
        """
        equations = []
        for i, p in enumerate(self.phases):
            if contains is not None and contains not in p.count:
                continue
            equations.append(f'{i}) {p.equation()}')
        return equations

    def diagram(self, U=None, pH=None):
        """Actual evaluation of the complete diagram

        Returns
        -------

        pour:
            The stability domains of the diagram on the pH vs. U grid.
            domains are represented by indexes (as integers)
            that map to Pourbaix.phases
            the target material is stable (index=-1)

        meta:
            The Pourbaix energy on the pH vs. U grid.

        text:
            The coordinates and phases information for
            text placement on the diagram.

        domains:
            ...

        """
        pour = np.zeros((len(U), len(pH)))
        meta = pour.copy()

        for i, u in enumerate(U):
            for j, p in enumerate(pH):
                meta[i, j], pour[i, j] = self._get_pourbaix_energy(u, p)

        # Identifying the region where the target material
        # is stable and updating the diagram accordingly
        where_stable = (meta <= 0)
        pour[where_stable] = -1

        text = []
        domains = [int(i) for i in np.unique(pour)]
        for phase_id in domains:
            if phase_id == -1:
                where = where_stable
                txt = {self.material.name: 1}
            else:
                where = (pour == phase_id)
                phase = self.phases[int(phase_id)]
                txt = phase.count
            x = np.dot(where.sum(1), U) / where.sum()
            y = np.dot(where.sum(0), pH) / where.sum()
            text.append((x, y, txt))

        return PourbaixDiagram(self, U, pH, pour, meta, text, domains)

    def get_phase_boundaries(self, phmin, phmax, umin, umax, domains, tol=1e-6):
        """Plane intersection method for finding
           the boundaries between phases seen in the final plot.

        Returns a list of tuples, each representing a single boundary,
        of the form ([[x1, x2], [y1, y2]], [id1, id2]), namely the
        x and y coordinates of the simplices connected by the boundary
        and the id's of the phases at each side of the boundary.
        """
        from collections import defaultdict
        from itertools import combinations

        # Planes identifying the diagram frame
        planes = [(np.array([0.0, 1.0, 0.0]), umin, 'bottom'),
                  (np.array([0.0, 1.0, 0.0]), umax, 'top'),
                  (np.array([1.0, 0.0, 0.0]), phmin, 'left'),
                  (np.array([1.0, 0.0, 0.0]), phmax, 'right')]

        # Planes associated with the stable domains of the diagram.
        # Given x=pH, y=U, z=E_pbx=-DeltaG, each plane has expression:
        # _vector[2]*x + _vector[1]*y + z = -_vector[0]
        # The region where the target material is stable
        # (id=-1, if present) is delimited by the xy plane (Epbx=0)
        for d in domains:
            if d == -1:
                plane = np.array([0.0, 0.0, 1.0])
                const = 0.0
            else:
                pvec = self.phases[d]._vector
                plane = np.array([pvec[2], pvec[1], 1])
                const = -pvec[0]
            planes.append((plane, const, d))

        # The simplices are found from the intersection points between
        # all possible plane triplets. If the z coordinate of the point
        # matches the corresponding pourbaix energy,
        # then the point is a simplex.
        simplices = []
        for (p1, c1, id1), (p2, c2, id2), (p3, c3, id3) in \
                combinations(planes, 3):
            A = np.vstack((p1, p2, p3))
            c = np.array([c1, c2, c3])
            ids = (id1, id2, id3)

            try:
                # triplets containing parallel planes raise a LinAlgError
                # and are automatically excluded.
                invA = np.linalg.inv(A)
            except np.linalg.LinAlgError:
                continue

            pt = np.dot(invA, c)
            Epbx = self._get_pourbaix_energy(pt[1], pt[0])[0]
            if pt[2] >= -tol and \
               np.isclose(Epbx, pt[2], rtol=0, atol=tol) and \
               phmin - tol <= pt[0] <= phmax + tol and \
               umin - tol <= pt[1] <= umax + tol:
                simplex = np.round(pt[:2], 3)
                simplices.append((simplex, ids))

        # The final segments to plot on the diagram are found from
        # the pairs of unique simplices that have two neighboring phases
        # in common, diagram frame excluded.
        duplicate_filter = defaultdict(int)

        def are_boundaries_there(comm):
            bounds = ['bottom', 'top', 'left', 'right']
            for c in comm:
                if c in bounds:
                    return True
            return False

        for (s1, id1), (s2, id2) in combinations(simplices, 2):

            # common neighboring phases, diagram frame excluded
            common = {*id1} & {*id2}

            simplices_are_distinct = not np.allclose(s1, s2, rtol=0, atol=tol)
            only_two_phases_in_common = (len(common) == 2)
            diagram_frame_excluded = not are_boundaries_there(common)

            # Filtering out duplicates
            if all([simplices_are_distinct, only_two_phases_in_common,
                    diagram_frame_excluded]):
                testarray = sorted([s1, s2], key=lambda s: (s[0], s[1]))
                testarray.append(sorted(common))
                testarray = np.array(testarray).flatten()
                duplicate_filter[tuple(testarray)] += 1

        segments = []
        for segment in duplicate_filter:
            coords, phases = np.split(np.array(segment), [4])
            segments.append((
                coords.reshape(2, 2).T,
                list(phases.astype(int))
            ))
        return segments


@dataclass
class PourbaixDiagram:
    pbx: Pourbaix
    U: np.ndarray
    pH: np.ndarray
    pour: np.ndarray
    meta: np.ndarray
    text: List[Tuple[float, float, str]]
    domains: List[int]

    def __post_init__(self):
        def _issorted(array):
            return all(np.diff(array) > 0)

        # We might as well require the input domains to be sorted:
        if not _issorted(self.U):
            raise ValueError('U must be sorted')

        if not _issorted(self.pH):
            raise ValueError('pH must be sorted')

    @property
    def pHrange(self):
        return (self.pH[0], self.pH[-1])

    @property
    def Urange(self):
        return (self.U[0], self.U[-1])

    def _draw_diagram_axes(
            self,
            cap,
            normalize,
            include_text,
            include_water,
            labeltype, cmap, *,
            ax):
        """Backend for drawing Pourbaix diagrams."""

        meta = self.meta.copy()
        if normalize:
            meta /= self.pbx.material.natoms
            cbarlabel = r'$\Delta G_{pbx}$ (eV/atom)'
        else:
            cbarlabel = r'$\Delta G_{pbx}$ (eV)'

        fig = ax.get_figure()
        extent = [*self.pHrange, *self.Urange]

        fig.subplots_adjust(
            left=0.1, right=0.97,
            top=0.97, bottom=0.14
        )

        if isinstance(cap, list):
            vmin = cap[0]
            vmax = cap[1]
        else:
            vmin = -cap
            vmax = cap

        colorplot = ax.imshow(
            meta, cmap=cmap,
            extent=extent,
            vmin=vmin, vmax=vmax,
            origin='lower', aspect='auto',
            interpolation='gaussian'
        )

        cbar = fig.colorbar(
            colorplot,
            ax=ax,
            pad=0.02
        )

        bounds = self.pbx.get_phase_boundaries(
            *self.pHrange, *self.Urange, self.domains
        )
        for coords, _ in bounds:
            ax.plot(coords[0], coords[1], '-', c='k', lw=1.0)

        if labeltype == 'numbers':
            add_numbers(ax, self.text)
        elif labeltype == 'phases':
            add_labels(ax, self.text)
        elif labeltype is None:
            pass
        else:
            raise ValueError("The provided label type doesn't exist")

        if include_water:
            add_redox_lines(ax, self.pH, self.pbx.reference, 'w')

        ax.set_xlim(*self.pHrange)
        ax.set_ylim(*self.Urange)
        ax.set_xlabel('pH', fontsize=22)
        ax.set_ylabel(r'$\it{U}$' + f' vs. {self.pbx.reference} (V)',
                      fontsize=22)
        ax.set_xticks(np.arange(self.pHrange[0], self.pHrange[1] + 1, 2))
        ax.set_yticks(np.arange(self.Urange[0], self.Urange[1] + 1, 1))
        ax.xaxis.set_tick_params(width=1.5, length=5)
        ax.yaxis.set_tick_params(width=1.5, length=5)
        ax.tick_params(axis='both', labelsize=22)

        ticks = np.linspace(vmin, vmax, num=5)
        cbar.set_ticks(ticks)
        cbar.set_ticklabels(ticks)
        cbar.outline.set_linewidth(1.5)
        cbar.ax.tick_params(labelsize=20, width=1.5, length=5)
        cbar.ax.set_ylabel(cbarlabel, fontsize=20)

        for axis in ['top', 'bottom', 'left', 'right']:
            ax.spines[axis].set_linewidth(1.5)

        if include_text:
            fig.subplots_adjust(right=0.75)
            add_phase_labels(fig, self.text, offset=0.05)
            return ax, cbar

        fig.tight_layout()
        return cbar

    def plot(self,
             cap=1.0,
             # figsize=[12, 6],
             normalize=True,
             include_text=True,
             include_water=False,
             labeltype='numbers',
             cmap="RdYlGn_r",
             filename=None,
             ax=None,
             show=False):
        """Plot a complete Pourbaix diagram.

        Keyword arguments
        -----------------

        Urange: list
            The potential range onto which to draw the diagram.

        pHrange: list
            The pH range onto which to draw the diagram.

        npoints: int
            The resolution of the diagram. Higher values
            mean higher resolution and thus higher compute times.

        cap: float/list
            If float, the limit (in both the positive and negative direction)
            of the Pourbaix energy colormap.
            If list, the first and second value determine the colormap limits.

        figsize: list
            The horizontal and vertical size of the graph.

        normalize: bool
            Normalize energies by the number of
            atoms in the target material unit formula.

        include_text: bool
            Report to the right of the diagram the main products
            associated with the stability domains.

        include_water: bool
            Include in the diagram the stability domain of water.

        labeltype: str/None
            The labeling style of the diagram domains. Options:
                'numbers': just add numbers associated with the
                           different phases, the latter shown
                           on the right if include_text=True.
                'phases':  Write the main products directly on the diagram.
                           These labels can be dragged around if the placement
                           is unsatisfactory. Redundant if include_text=True.
                 None:     Don't draw any labels.

        filename: str/None
            If passed as a string, the figure will be saved with that name.

        show: bool
            Spawn a window showing the diagram.

        """
        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.gca()

        fig = ax.get_figure()

        self._draw_diagram_axes(
            cap,
            normalize,
            include_text,
            include_water,
            labeltype, cmap,
            ax=ax)

        if filename is not None:
            fig.savefig(filename)

        if show:
            plt.show()



import numpy as np
import pytest
from pytest import mark

from ase import Atoms


@mark.calculator_lite
def test_update_neighbor_parameters(KIM):
    """
    Check that the neighbor lists are updated properly when model parameters
    are updated. This is done by instantiating the calculator for a specific
    Lennard-Jones (LJ) potential, included with the KIM API, for Mo-Mo
    interactions.  First, an equally spaced collinear trimer is constructed
    whose separation distances are just over half the cutoff distance
    associated with Mo so that only nearest-neighbor interactions occur (the
    end atoms do not interact with one another) and the energy is computed.
    Next, the model's cutoff parameter associated with Mo is increased to a
    distance greater than the distance between the end atoms, so that their
    interaction will produce a (fairly small) non-zero value.  The energy is
    computed once again and it is verified that this energy differs
    significantly from the initial energy computed, thus implying that the
    neighbor lists of the end atoms must have been updated so as to contain one
    another.
    """

    # Create LJ calculator and disable energy shifting.  Otherwise, the energy
    # would always change when the cutoff changes, even if the neighbor lists
    # remained the same.
    calc = KIM(
        "LennardJones612_UniversalShifted__MO_959249795837_003",
        options={"neigh_skin_ratio": 0.0},
    )
    calc.set_parameters(shift=[0, 0])

    # Set all "cutoffs" parameters in the calculator to zero, except for the
    # one used for Mo.  This is necessary because we want to control the
    # influence distance, which is the maximum of all of the cutoffs.
    Mo_cutoff_index = 4879
    Mo_cutoff = calc.get_parameters(cutoffs=Mo_cutoff_index)["cutoffs"][1]
    cutoffs_extent = calc.parameters_metadata()["cutoffs"]["extent"]
    calc.set_parameters(
        cutoffs=[
            list(
                range(cutoffs_extent)),
            [0.0] *
            cutoffs_extent])
    calc.set_parameters(cutoffs=[Mo_cutoff_index, Mo_cutoff])

    # Create trimer such that nearest neighbor interactions occur:  each of the
    # end atoms see the middle atom, and the middle atom sees both end atoms,
    # but the end atoms do not interact with one another
    nearest_neighbor_separation = np.random.RandomState(11).uniform(
        0.6 * Mo_cutoff, 0.65 * Mo_cutoff
    )
    pos = [
        [0, 0, 0],
        [0, 0, nearest_neighbor_separation],
        [0, 0, 2 * nearest_neighbor_separation],
    ]
    trimer = Atoms("Mo" * 3, positions=pos)
    trimer.calc = calc

    eng_orig = trimer.get_potential_energy()

    # Update the cutoff parameter so that the end atoms will interact with one
    # another
    long_cutoff = 1.1 * \
        np.linalg.norm(np.array(pos[2][:]) - np.array(pos[0][:]))
    calc.set_parameters(cutoffs=[Mo_cutoff_index, long_cutoff])

    # Energy of the trimer after modifying cutoff
    eng_modified = trimer.get_potential_energy()

    # Check if the far atom is excluded when the original cutoff is used
    assert eng_modified != pytest.approx(eng_orig, rel=1e-4)



import re

import numpy as np

from ase.units import Angstrom, Bohr, Debye, Hartree, eV


class OctopusIOError(IOError):
    pass  # Cannot find output files


def read_eigenvalues_file(fd):
    unit = None

    for line in fd:
        m = re.match(r'Eigenvalues\s*\[(.+?)\]', line)
        if m is not None:
            unit = m.group(1)
            break
    line = next(fd)
    assert line.strip().startswith('#st'), line

    # fermilevel = None
    kpts = []
    eigs = []
    occs = []

    for line in fd:
        m = re.match(r'#k.*?\(\s*(.+?),\s*(.+?),\s*(.+?)\)', line)
        if m:
            k = m.group(1, 2, 3)
            kpts.append(np.array(k, float))
            eigs.append({})
            occs.append({})
        else:
            m = re.match(r'\s*\d+\s*(\S+)\s*(\S+)\s*(\S+)', line)
            if m is None:
                m = re.match(r'Fermi energy\s*=\s*(\S+)\s*', line)
                assert m is not None
                # We can also return the fermilevel but so far we just read
                # it from the static/info instead.
                # fermilevel = float(m.group(1))
            else:
                spin, eig, occ = m.group(1, 2, 3)

                if not eigs:
                    # Only initialized if kpoint header was written
                    eigs.append({})
                    occs.append({})

                eigs[-1].setdefault(spin, []).append(float(eig))
                occs[-1].setdefault(spin, []).append(float(occ))

    nkpts = len(kpts)
    nspins = len(eigs[0])
    nbands = len(eigs[0][spin])

    kptsarr = np.array(kpts, float)
    eigsarr = np.empty((nkpts, nspins, nbands))
    occsarr = np.empty((nkpts, nspins, nbands))

    arrs = [eigsarr, occsarr]

    for arr in arrs:
        arr.fill(np.nan)

    for k in range(nkpts):
        for arr, lst in [(eigsarr, eigs), (occsarr, occs)]:
            arr[k, :, :] = [lst[k][sp] for sp
                            in (['--'] if nspins == 1 else ['up', 'dn'])]

    for arr in arrs:
        assert not np.isnan(arr).any()

    eigsarr *= {'H': Hartree, 'eV': eV}[unit]
    return kptsarr, eigsarr, occsarr


def read_static_info_stress(fd):
    stress_cv = np.empty((3, 3))

    headers = next(fd)
    assert headers.strip().startswith('T_{ij}')
    for i in range(3):
        line = next(fd)
        tokens = line.split()
        vec = np.array(tokens[1:4]).astype(float)
        stress_cv[i] = vec
    return stress_cv


def read_static_info_kpoints(fd):
    for line in fd:
        if line.startswith('List of k-points'):
            break

    tokens = next(fd).split()
    assert tokens == ['ik', 'k_x', 'k_y', 'k_z', 'Weight']
    bar = next(fd)
    assert bar.startswith('---')

    kpts = []
    weights = []

    for line in fd:
        # Format:        index   kx      ky      kz     weight
        m = re.match(r'\s*\d+\s*(\S+)\s*(\S+)\s*(\S+)\s*(\S+)', line)
        if m is None:
            break
        kxyz = m.group(1, 2, 3)
        weight = m.group(4)
        kpts.append(kxyz)
        weights.append(weight)

    ibz_kpoints = np.array(kpts, float)
    kpoint_weights = np.array(weights, float)
    return dict(ibz_kpoints=ibz_kpoints, kpoint_weights=kpoint_weights)


def read_static_info_eigenvalues(fd, energy_unit):

    values_sknx = {}

    nbands = 0
    fermilevel = None
    for line in fd:
        line = line.strip()
        if line.startswith('#'):
            continue
        if not line[:1].isdigit():
            m = re.match(r'Fermi energy\s*=\s*(\S+)', line)
            if m is not None:
                fermilevel = float(m.group(1)) * energy_unit
            break

        tokens = line.split()
        nbands = max(nbands, int(tokens[0]))
        energy = float(tokens[2]) * energy_unit
        occupation = float(tokens[3])
        values_sknx.setdefault(tokens[1], []).append((energy, occupation))

    nspins = len(values_sknx)
    if nspins == 1:
        val = [values_sknx['--']]
    else:
        val = [values_sknx['up'], values_sknx['dn']]
    val = np.array(val, float)
    nkpts, remainder = divmod(len(val[0]), nbands)
    assert remainder == 0

    eps_skn = val[:, :, 0].reshape(nspins, nkpts, nbands)
    occ_skn = val[:, :, 1].reshape(nspins, nkpts, nbands)
    eps_skn = eps_skn.transpose(1, 0, 2).copy()
    occ_skn = occ_skn.transpose(1, 0, 2).copy()
    assert eps_skn.flags.contiguous
    d = dict(nspins=nspins,
             nkpts=nkpts,
             nbands=nbands,
             eigenvalues=eps_skn,
             occupations=occ_skn)
    if fermilevel is not None:
        d.update(fermi_level=fermilevel)
    return d


def read_static_info_energy(fd, energy_unit):
    def get(name):
        for line in fd:
            if line.strip().startswith(name):
                return float(line.split('=')[-1].strip()) * energy_unit
    return dict(energy=get('Total'), free_energy=get('Free'))


def read_static_info(fd):
    results = {}

    def get_energy_unit(line):  # Convert "title [unit]": ---> unit
        return {'[eV]': eV, '[H]': Hartree}[line.split()[1].rstrip(':')]

    for line in fd:
        if line.strip('*').strip().startswith('Brillouin zone'):
            results.update(read_static_info_kpoints(fd))
        elif line.startswith('Eigenvalues ['):
            unit = get_energy_unit(line)
            results.update(read_static_info_eigenvalues(fd, unit))
        elif line.startswith('Energy ['):
            unit = get_energy_unit(line)
            results.update(read_static_info_energy(fd, unit))
        elif line.startswith('Total stress tensor ['):
            if '[H/b^3]' in line:
                stress = read_static_info_stress(fd)
                stress *= Hartree / Bohr**3
                results.update(stress=stress)
        elif line.startswith('Total Magnetic Moment'):
            line = next(fd)
            values = line.split()
            results['magmom'] = float(values[-1])
            line = next(fd)
            assert line.startswith('Local Magnetic Moments')
            line = next(fd)
            assert line.split() == ['Ion', 'mz']
            # Reading  Local Magnetic Moments
            magmoms = []
            for line in fd:
                if line == '\n':
                    break  # there is no more thing to search for
                line = line.replace('\n', ' ')
                values = line.split()
                magmoms.append(float(values[-1]))
            results['magmoms'] = np.array(magmoms)
        elif line.startswith('Dipole'):
            assert line.split()[-1] == '[Debye]'
            dipole = [float(next(fd).split()[-1]) for i in range(3)]
            results['dipole'] = np.array(dipole) * Debye
        elif line.startswith('Forces'):
            forceunitspec = line.split()[-1]
            forceunit = {'[eV/A]': eV / Angstrom,
                         '[H/b]': Hartree / Bohr}[forceunitspec]
            forces = []
            line = next(fd)
            assert line.strip().startswith('Ion')
            for line in fd:
                if line.strip().startswith('---'):
                    break
                tokens = line.split()[-3:]
                forces.append([float(f) for f in tokens])
            results['forces'] = np.array(forces) * forceunit

    return results



import numpy as np

from ase.build import fcc111
from ase.constraints import FixAtoms
from ase.ga import set_raw_score
from ase.ga.data import DataConnection, PrepareDB
from ase.ga.startgenerator import StartGenerator
from ase.ga.utilities import closest_distances_generator

db_file = 'gadb_logics_test.db'


def test_database_logic(seed, testdir):

    # set up the random number generator
    rng = np.random.RandomState(seed)

    slab = fcc111('Au', size=(4, 4, 2), vacuum=10.0, orthogonal=True)
    slab.set_constraint(FixAtoms(mask=slab.positions[:, 2] <= 10.))

    # define the volume in which the adsorbed cluster is optimized
    # the volume is defined by a corner position (p0)
    # and three spanning vectors (v1, v2, v3)
    pos = slab.get_positions()
    cell = slab.get_cell()
    p0 = np.array([0., 0., max(pos[:, 2]) + 2.])
    v1 = cell[0, :] * 0.8
    v2 = cell[1, :] * 0.8
    v3 = cell[2, :]
    v3[2] = 3.

    # define the closest distance between two atoms of a given species
    blmin = closest_distances_generator(atom_numbers=[47, 79],
                                        ratio_of_covalent_radii=0.7)

    # Define the composition of the atoms to optimize
    atom_numbers = 2 * [47] + 2 * [79]

    # create the starting population
    sg = StartGenerator(slab=slab,
                        blocks=atom_numbers,
                        blmin=blmin,
                        box_to_place_in=[p0, [v1, v2, v3]],
                        rng=rng)

    # generate the starting population
    starting_population = [sg.get_new_candidate() for _ in range(20)]

    d = PrepareDB(db_file_name=db_file,
                  simulation_cell=slab,
                  stoichiometry=atom_numbers)

    for a in starting_population:
        d.add_unrelaxed_candidate(a)

    # and now for the actual test
    dc = DataConnection(db_file)

    dc.get_slab()
    dc.get_atom_numbers_to_optimize()

    assert dc.get_number_of_unrelaxed_candidates() == 20

    a1 = dc.get_an_unrelaxed_candidate()
    dc.mark_as_queued(a1)

    assert dc.get_number_of_unrelaxed_candidates() == 19
    assert len(dc.get_all_candidates_in_queue()) == 1

    set_raw_score(a1, 0.0)
    dc.add_relaxed_step(a1)

    assert dc.get_number_of_unrelaxed_candidates() == 19
    assert len(dc.get_all_candidates_in_queue()) == 0

    assert len(dc.get_all_relaxed_candidates()) == 1

    a2 = dc.get_an_unrelaxed_candidate()
    dc.mark_as_queued(a2)
    confid = a2.info['confid']
    assert dc.get_all_candidates_in_queue()[0] == confid

    dc.remove_from_queue(confid)
    assert len(dc.get_all_candidates_in_queue()) == 0



from math import pi

import numpy as np

from ase.atoms import Atoms
from ase.calculators.calculator import Calculator, kpts2ndarray
from ase.calculators.fd import calculate_numerical_forces
from ase.units import Bohr, Ha


def make_test_dft_calculation():
    a = b = 2.0
    c = 6.0
    atoms = Atoms(positions=[(0, 0, c / 2)],
                  symbols='H',
                  pbc=(1, 1, 0),
                  cell=(a, b, c),
                  calculator=TestCalculator())
    return atoms


class TestCalculator:
    def __init__(self, nk=8):
        assert nk % 2 == 0
        bzk = []
        weights = []
        ibzk = []
        w = 1.0 / nk**2
        for i in range(-nk + 1, nk, 2):
            for j in range(-nk + 1, nk, 2):
                k = (0.5 * i / nk, 0.5 * j / nk, 0)
                bzk.append(k)
                if i >= j > 0:
                    ibzk.append(k)
                    if i == j:
                        weights.append(4 * w)
                    else:
                        weights.append(8 * w)
        assert abs(sum(weights) - 1.0) < 1e-12
        self.bzk = np.array(bzk)
        self.ibzk = np.array(ibzk)
        self.weights = np.array(weights)

        # Calculate eigenvalues and wave functions:
        self.init()

    def init(self):
        nibzk = len(self.weights)
        nbands = 1

        V = -1.0
        self.eps = 2 * V * (np.cos(2 * pi * self.ibzk[:, 0]) +
                            np.cos(2 * pi * self.ibzk[:, 1]))
        self.eps.shape = (nibzk, nbands)

        self.psi = np.zeros((nibzk, 20, 20, 60), complex)
        phi = np.empty((2, 2, 20, 20, 60))
        z = np.linspace(-1.5, 1.5, 60, endpoint=False)
        for i in range(2):
            x = np.linspace(0, 1, 20, endpoint=False) - i
            for j in range(2):
                y = np.linspace(0, 1, 20, endpoint=False) - j
                r = (((x[:, None]**2 +
                       y**2)[:, :, None] +
                      z**2)**0.5).clip(0, 1)
                phi = 1.0 - r**2 * (3.0 - 2.0 * r)
                phase = np.exp(pi * 2j * np.dot(self.ibzk, (i, j, 0)))
                self.psi += phase[:, None, None, None] * phi

    def get_pseudo_wave_function(self, band=0, kpt=0, spin=0):
        assert spin == 0 and band == 0
        return self.psi[kpt]

    def get_eigenvalues(self, kpt=0, spin=0):
        assert spin == 0
        return self.eps[kpt]

    def get_number_of_bands(self):
        return 1

    def get_k_point_weights(self):
        return self.weights

    def get_number_of_spins(self):
        return 1

    def get_fermi_level(self):
        return 0.0

    def get_pseudo_density(self):
        n = 0.0
        for w, eps, psi in zip(self.weights, self.eps[:, 0], self.psi):
            if eps >= 0.0:
                continue
            n += w * (psi * psi.conj()).real

        n[1:] += n[:0:-1].copy()
        n[:, 1:] += n[:, :0:-1].copy()
        n += n.transpose((1, 0, 2)).copy()
        n /= 8
        return n


class TestPotential(Calculator):
    implemented_properties = ['energy', 'forces']

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        E = 0.0
        R = atoms.positions
        F = np.zeros_like(R)
        for a, r in enumerate(R):
            D = R - r
            d = (D**2).sum(1)**0.5
            x = d - 1.0
            E += np.vdot(x, x)
            d[a] = 1
            F -= (x / d)[:, None] * D
        energy = 0.25 * E
        self.results = {'energy': energy, 'forces': F}


class FreeElectrons(Calculator):
    """Free-electron band calculator.

    Parameters:

    nvalence: int
        Number of electrons
    kpts: dict
        K-point specification.

    Example:
    >>> from ase.calculators.test import FreeElectrons
    >>> calc = FreeElectrons(nvalence=1, kpts={'path': 'GXL'})
    """

    implemented_properties = ['energy']
    default_parameters = {'kpts': np.zeros((1, 3)),
                          'nvalence': 0.0,
                          'nbands': 20,
                          'gridsize': 7}

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms)
        self.kpts = kpts2ndarray(self.parameters.kpts, atoms)
        icell = atoms.cell.reciprocal() * 2 * np.pi * Bohr
        n = self.parameters.gridsize
        offsets = np.indices((n, n, n)).T.reshape((n**3, 1, 3)) - n // 2
        eps = 0.5 * (np.dot(self.kpts + offsets, icell)**2).sum(2).T
        eps.sort()
        self.eigenvalues = eps[:, :self.parameters.nbands] * Ha
        self.results = {'energy': 0.0}

    def get_eigenvalues(self, kpt, spin=0):
        assert spin == 0
        return self.eigenvalues[kpt].copy()

    def get_fermi_level(self):
        v = self.atoms.get_volume() / Bohr**3
        kF = (self.parameters.nvalence / v * 3 * np.pi**2)**(1 / 3)
        return 0.5 * kF**2 * Ha

    def get_ibz_k_points(self):
        return self.kpts.copy()

    def get_number_of_spins(self):
        return 1


def gradient_test(atoms, indices=None):
    """
    Use numeric_force to compare analytical and numerical forces on atoms

    If indices is None, test is done on all atoms.
    """
    if indices is None:
        indices = range(len(atoms))
    f = atoms.get_forces()[indices]
    print('{:>16} {:>20}'.format('eps', 'max(abs(df))'))
    for eps in np.logspace(-1, -8, 8):
        fn = calculate_numerical_forces(atoms, eps, indices)
        print(f'{eps:16.12f} {abs(fn - f).max():20.12f}')
    return f, fn



import io
import re
import warnings
from collections import defaultdict
from typing import Any, Dict, List, Optional

import numpy as np

from ase import Atoms, units
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms, FixCartesian, FixConstraint
from ase.io import ParseError
from ase.utils import reader, string2index


@reader
def read_castep_castep(fd, index=-1):
    """Read a .castep file and returns an Atoms object.

    The calculator information will be stored in the calc attribute.

    Notes
    -----
    This routine will return an atom ordering as found within the castep file.
    This means that the species will be ordered by ascending atomic numbers.
    The atoms witin a species are ordered as given in the original cell file.

    """
    # look for last result, if several CASTEP run are appended
    line_start, line_end, end_found = _find_last_record(fd)
    if not end_found:
        raise ParseError(f'No regular end found in {fd.name} file.')

    # These variables are finally assigned to `SinglePointCalculator`
    # for backward compatibility with the `Castep` calculator.
    cut_off_energy = None
    kpoints = None
    total_time = None
    peak_memory = None

    # jump back to the beginning to the last record
    fd.seek(0)
    for i, line in enumerate(fd):
        if i == line_start:
            break

    # read header
    parameters_header = _read_header(fd)
    if 'cut_off_energy' in parameters_header:
        cut_off_energy = parameters_header['cut_off_energy']
        if 'basis_precision' in parameters_header:
            del parameters_header['cut_off_energy']  # avoid conflict

    markers_new_iteration = [
        'BFGS: starting iteration',
        'BFGS: improving iteration',
        'Starting MD iteration',
    ]

    images = []

    results = {}
    constraints = []
    species_pot = []
    castep_warnings = []
    for i, line in enumerate(fd):

        if i > line_end:
            break

        if 'Number of kpoints used' in line:
            kpoints = int(line.split('=')[-1].strip())
        elif 'Unit Cell' in line:
            lattice_real = _read_unit_cell(fd)
        elif 'Cell Contents' in line:
            for line in fd:
                if 'Total number of ions in cell' in line:
                    n_atoms = int(line.split()[7])
                if 'Total number of species in cell' in line:
                    int(line.split()[7])
                fields = line.split()
                if len(fields) == 0:
                    break
        elif 'Fractional coordinates of atoms' in line:
            species, custom_species, positions_frac = \
                _read_fractional_coordinates(fd, n_atoms)
        elif 'Files used for pseudopotentials' in line:
            for line in fd:
                line = fd.readline()
                if 'Pseudopotential generated on-the-fly' in line:
                    continue
                fields = line.split()
                if len(fields) == 2:
                    species_pot.append(fields)
                else:
                    break
        elif 'k-Points For BZ Sampling' in line:
            # TODO: generalize for non-Monkhorst Pack case
            # (i.e. kpoint lists) -
            # kpoints_offset cannot be read this way and
            # is hence always set to None
            for line in fd:
                if not line.strip():
                    break
                if 'MP grid size for SCF calculation' in line:
                    # kpoints =  ' '.join(line.split()[-3:])
                    # self.kpoints_mp_grid = kpoints
                    # self.kpoints_mp_offset = '0. 0. 0.'
                    # not set here anymore because otherwise
                    # two calculator objects go out of sync
                    # after each calculation triggering unnecessary
                    # recalculation
                    break

        elif 'Final energy' in line:
            key = 'energy_without_dispersion_correction'
            results[key] = float(line.split()[-2])
        elif 'Final free energy' in line:
            key = 'free_energy_without_dispersion_correction'
            results[key] = float(line.split()[-2])
        elif 'NB est. 0K energy' in line:
            key = 'energy_zero_without_dispersion_correction'
            results[key] = float(line.split()[-2])

        # Add support for dispersion correction
        # filtering due to SEDC is done in get_potential_energy
        elif 'Dispersion corrected final energy' in line:
            key = 'energy_with_dispersion_correlation'
            results[key] = float(line.split()[-2])
        elif 'Dispersion corrected final free energy' in line:
            key = 'free_energy_with_dispersion_correlation'
            results[key] = float(line.split()[-2])
        elif 'NB dispersion corrected est. 0K energy' in line:
            key = 'energy_zero_with_dispersion_correlation'
            results[key] = float(line.split()[-2])

        # check if we had a finite basis set correction
        elif 'Total energy corrected for finite basis set' in line:
            key = 'energy_with_finite_basis_set_correction'
            results[key] = float(line.split()[-2])

        # ******************** Forces *********************
        # ************** Symmetrised Forces ***************
        # ******************** Constrained Forces ********************
        # ******************* Unconstrained Forces *******************
        elif re.search(r'\**.* Forces \**', line):
            forces, constraints = _read_forces(fd, n_atoms)
            results['forces'] = np.array(forces)

        # ***************** Stress Tensor *****************
        # *********** Symmetrised Stress Tensor ***********
        elif re.search(r'\**.* Stress Tensor \**', line):
            results.update(_read_stress(fd))

        elif any(_ in line for _ in markers_new_iteration):
            _add_atoms(
                images,
                lattice_real,
                species,
                custom_species,
                positions_frac,
                constraints,
                results,
            )
            # reset for the next step
            lattice_real = None
            species = None
            positions_frac = None
            results = {}
            constraints = []

        # extract info from the Mulliken analysis
        elif 'Atomic Populations' in line:
            results.update(_read_mulliken_charges(fd))

        # extract detailed Hirshfeld analysis (iprint > 1)
        elif 'Hirshfeld total electronic charge (e)' in line:
            results.update(_read_hirshfeld_details(fd, n_atoms))

        elif 'Hirshfeld Analysis' in line:
            results.update(_read_hirshfeld_charges(fd))

        # There is actually no good reason to get out of the loop
        # already at this point... or do I miss something?
        # elif 'BFGS: Final Configuration:' in line:
        #    break
        elif 'warn' in line.lower():
            castep_warnings.append(line)

        # fetch some last info
        elif 'Total time' in line:
            pattern = r'.*=\s*([\d\.]+) s'
            total_time = float(re.search(pattern, line).group(1))

        elif 'Peak Memory Use' in line:
            pattern = r'.*=\s*([\d]+) kB'
            peak_memory = int(re.search(pattern, line).group(1))

    # add the last image
    _add_atoms(
        images,
        lattice_real,
        species,
        custom_species,
        positions_frac,
        constraints,
        results,
    )

    for atoms in images:
        # these variables are temporarily assigned to `SinglePointCalculator`
        # to be assigned to the `Castep` calculator for backward compatibility
        atoms.calc._cut_off_energy = cut_off_energy
        atoms.calc._kpoints = kpoints
        atoms.calc._species_pot = species_pot
        atoms.calc._total_time = total_time
        atoms.calc._peak_memory = peak_memory
        atoms.calc._parameters_header = parameters_header

    if castep_warnings:
        warnings.warn('WARNING: .castep file contains warnings')
        for warning in castep_warnings:
            warnings.warn(warning)

    if isinstance(index, str):
        index = string2index(index)

    return images[index]


def _find_last_record(fd):
    """Find the last record of the .castep file.

    Returns
    -------
    start : int
        Line number of the first line of the last record.
    end : int
        Line number of the last line of the last record.
    end_found : bool
        True if the .castep file ends as expected.

    """
    start = -1
    for i, line in enumerate(fd):
        if (('Welcome' in line or 'Materials Studio' in line)
                and 'CASTEP' in line):
            start = i

    if start < 0:
        warnings.warn(
            f'Could not find CASTEP label in result file: {fd.name}.'
            ' Are you sure this is a .castep file?'
        )
        return None

    # search for regular end of file
    end_found = False
    # start to search from record beginning from the back
    # and see if
    end = -1
    fd.seek(0)
    for i, line in enumerate(fd):
        if i < start:
            continue
        if 'Finalisation time   =' in line:
            end_found = True
            end = i
            break

    return (start, end, end_found)


def _read_header(out: io.TextIOBase):
    """Read the header blocks from a .castep file.

    Returns
    -------
    parameters : dict
        Dictionary storing keys and values of a .param file.
    """
    def _parse_on_off(_: str):
        return {'on': True, 'off': False}[_]

    read_title = False
    parameters: Dict[str, Any] = {}
    for line in out:
        if len(line) == 0:  # end of file
            break
        if re.search(r'^\s*\*+$', line) and read_title:  # end of header
            break

        if re.search(r'\**.* Title \**', line):
            read_title = True

        # General Parameters

        elif 'output verbosity' in line:
            parameters['iprint'] = int(line.split()[-1][1])
        elif re.match(r'\stype of calculation\s*:', line):
            parameters['task'] = {
                'single point energy': 'SinglePoint',
                'geometry optimization': 'GeometryOptimization',
                'band structure': 'BandStructure',
                'molecular dynamics': 'MolecularDynamics',
                'optical properties': 'Optics',
                'phonon calculation': 'Phonon',
                'E-field calculation': 'Efield',
                'Phonon followed by E-field': 'Phonon+Efield',
                'transition state search': 'TransitionStateSearch',
                'Magnetic Resonance': 'MagRes',
                'Core level spectra': 'Elnes',
                'Electronic Spectroscopy': 'ElectronicSpectroscopy',
            }[line.split(':')[-1].strip()]
        elif 'stress calculation' in line:
            parameters['calculate_stress'] = _parse_on_off(line.split()[-1])
        elif 'calculation limited to maximum' in line:
            parameters['run_time'] = float(line.split()[-2])
        elif re.match(r'\soptimization strategy\s*:', line):
            parameters['opt_strategy'] = {
                'maximize speed(+++)': 'Speed',
                'minimize memory(---)': 'Memory',
                'balance speed and memory': 'Default',
            }[line.split(':')[-1].strip()]

        # Exchange-Correlation Parameters

        elif re.match(r'\susing functional\s*:', line):
            parameters['xc_functional'] = {
                'Local Density Approximation': 'LDA',
                'Perdew Wang (1991)': 'PW91',
                'Perdew Burke Ernzerhof': 'PBE',
                'revised Perdew Burke Ernzerhof': 'RPBE',
                'PBE with Wu-Cohen exchange': 'WC',
                'PBE for solids (2008)': 'PBESOL',
                'Hartree-Fock': 'HF',
                'Hartree-Fock +': 'HF-LDA',
                'Screened Hartree-Fock': 'sX',
                'Screened Hartree-Fock + ': 'sX-LDA',
                'hybrid PBE0': 'PBE0',
                'hybrid B3LYP': 'B3LYP',
                'hybrid HSE03': 'HSE03',
                'hybrid HSE06': 'HSE06',
                'RSCAN': 'RSCAN',
            }[line.split(':')[-1].strip()]
        elif 'DFT+D: Semi-empirical dispersion correction' in line:
            parameters['sedc_apply'] = _parse_on_off(line.split()[-1])
        elif 'SEDC with' in line:
            parameters['sedc_scheme'] = {
                'OBS correction scheme': 'OBS',
                'G06 correction scheme': 'G06',
                'D3 correction scheme': 'D3',
                'D3(BJ) correction scheme': 'D3-BJ',
                'D4 correction scheme': 'D4',
                'JCHS correction scheme': 'JCHS',
                'TS correction scheme': 'TS',
                'TSsurf correction scheme': 'TSSURF',
                'TS+SCS correction scheme': 'TSSCS',
                'aperiodic TS+SCS correction scheme': 'ATSSCS',
                'aperiodic MBD@SCS method': 'AMBD',
                'MBD@SCS method': 'MBD',
                'aperiodic MBD@rsSCS method': 'AMBD*',
                'MBD@rsSCS method': 'MBD*',
                'XDM correction scheme': 'XDM',
            }[line.split(':')[-1].strip()]

        # Basis Set Parameters

        elif 'basis set accuracy' in line:
            parameters['basis_precision'] = line.split()[-1]
        elif 'plane wave basis set cut-off' in line:
            parameters['cut_off_energy'] = float(line.split()[-2])
        elif re.match(r'\sfinite basis set correction\s*:', line):
            parameters['finite_basis_corr'] = {
                'none': 0,
                'manual': 1,
                'automatic': 2,
            }[line.split()[-1]]

        # Electronic Parameters

        elif 'treating system as spin-polarized' in line:
            parameters['spin_polarized'] = True

        # Electronic Minimization Parameters

        elif 'Treating system as non-metallic' in line:
            parameters['fix_occupancy'] = True
        elif 'total energy / atom convergence tol.' in line:
            parameters['elec_energy_tol'] = float(line.split()[-2])
        elif 'convergence tolerance window' in line:
            parameters['elec_convergence_win'] = int(line.split()[-2])
        elif 'max. number of SCF cycles:' in line:
            parameters['max_scf_cycles'] = float(line.split()[-1])
        elif 'dump wavefunctions every' in line:
            parameters['num_dump_cycles'] = float(line.split()[-3])

        # Density Mixing Parameters

        elif 'density-mixing scheme' in line:
            parameters['mixing_scheme'] = line.split()[-1]

    return parameters


def _read_unit_cell(out: io.TextIOBase):
    """Read a Unit Cell block from a .castep file."""
    for line in out:
        fields = line.split()
        if len(fields) == 6:
            break
    lattice_real = []
    for _ in range(3):
        lattice_real.append([float(f) for f in fields[0:3]])
        line = out.readline()
        fields = line.split()
    return np.array(lattice_real)


def _read_forces(out: io.TextIOBase, n_atoms: int):
    """Read a block for atomic forces from a .castep file."""
    constraints: List[FixConstraint] = []
    forces = []
    for line in out:
        fields = line.split()
        if len(fields) == 7:
            break
    for n in range(n_atoms):
        consd = np.array([0, 0, 0])
        fxyz = [0.0, 0.0, 0.0]
        for i, force_component in enumerate(fields[-4:-1]):
            if force_component.count("(cons'd)") > 0:
                consd[i] = 1
            # remove constraint labels in force components
            fxyz[i] = float(force_component.replace("(cons'd)", ''))
        if consd.all():
            constraints.append(FixAtoms(n))
        elif consd.any():
            constraints.append(FixCartesian(n, consd))
        forces.append(fxyz)
        line = out.readline()
        fields = line.split()
    return forces, constraints


def _read_fractional_coordinates(out: io.TextIOBase, n_atoms: int):
    """Read fractional coordinates from a .castep file."""
    species: List[str] = []
    custom_species: Optional[List[str]] = None  # A CASTEP special thing
    positions_frac: List[List[float]] = []
    for line in out:
        fields = line.split()
        if len(fields) == 7:
            break
    for _ in range(n_atoms):
        spec_custom = fields[1].split(':', 1)
        elem = spec_custom[0]
        if len(spec_custom) > 1 and custom_species is None:
            # Add it to the custom info!
            custom_species = list(species)
        species.append(elem)
        if custom_species is not None:
            custom_species.append(fields[1])
        positions_frac.append([float(s) for s in fields[3:6]])
        line = out.readline()
        fields = line.split()
    return species, custom_species, positions_frac


def _read_stress(out: io.TextIOBase):
    """Read a block for the stress tensor from a .castep file."""
    for line in out:
        fields = line.split()
        if len(fields) == 6:
            break
    results = {}
    stress = []
    for _ in range(3):
        stress.append([float(s) for s in fields[2:5]])
        line = out.readline()
        fields = line.split()
    # stress in .castep file is given in GPa
    results['stress'] = np.array(stress) * units.GPa
    results['stress'] = results['stress'].reshape(9)[[0, 4, 8, 5, 2, 1]]
    line = out.readline()
    if "Pressure:" in line:
        results['pressure'] = float(
            line.split()[-2]) * units.GPa  # type: ignore[assignment]
    return results


def _add_atoms(
    images,
    lattice_real,
    species,
    custom_species,
    positions_frac,
    constraints,
    results,
):
    # If all the lattice parameters are fixed,
    # the `Unit Cell` block in the .castep file is not printed
    # in every ionic step.
    # The lattice paramters are therefore taken from the last step.
    # This happens:
    # - `GeometryOptimization`: `FIX_ALL_CELL : TRUE`
    # - `MolecularDynamics`: `MD_ENSEMBLE : NVE or NVT`
    if lattice_real is None:
        lattice_real = images[-1].cell.copy()

    # for highly symmetric systems (where essentially only the
    # stress is optimized, but the atomic positions) positions
    # are only printed once.
    if species is None:
        species = images[-1].symbols
    if positions_frac is None:
        positions_frac = images[-1].get_scaled_positions()

    _set_energy_and_free_energy(results)

    atoms = Atoms(
        species,
        cell=lattice_real,
        constraint=constraints,
        pbc=True,
        scaled_positions=positions_frac,
    )
    if custom_species is not None:
        atoms.new_array(
            'castep_custom_species',
            np.array(custom_species),
        )

    atoms.calc = SinglePointCalculator(atoms)
    atoms.calc.results = results

    images.append(atoms)


def _read_mulliken_charges(out: io.TextIOBase):
    """Read a block for Mulliken charges from a .castep file."""
    for i in range(3):
        line = out.readline()
        if i == 1:
            spin_polarized = 'Spin' in line
    results = defaultdict(list)
    for line in out:
        fields = line.split()
        if len(fields) == 1:
            break
        if spin_polarized:
            if len(fields) != 7:  # due to CASTEP 18 outformat changes
                results['charges'].append(float(fields[-2]))
                results['magmoms'].append(float(fields[-1]))
        else:
            results['charges'].append(float(fields[-1]))
    return {k: np.array(v) for k, v in results.items()}


def _read_hirshfeld_details(out: io.TextIOBase, n_atoms: int):
    """Read the Hirshfeld analysis when iprint > 1 from a .castep file."""
    results = defaultdict(list)
    for _ in range(n_atoms):
        for line in out:
            if line.strip() == '':
                break  # end for each atom
            if 'Hirshfeld / free atomic volume :' in line:
                line = out.readline()
                fields = line.split()
                results['hirshfeld_volume_ratios'].append(float(fields[0]))
    return {k: np.array(v) for k, v in results.items()}


def _read_hirshfeld_charges(out: io.TextIOBase):
    """Read a block for Hirshfeld charges from a .castep file."""
    for i in range(3):
        line = out.readline()
        if i == 1:
            spin_polarized = 'Spin' in line
    results = defaultdict(list)
    for line in out:
        fields = line.split()
        if len(fields) == 1:
            break
        if spin_polarized:
            results['hirshfeld_charges'].append(float(fields[-2]))
            results['hirshfeld_magmoms'].append(float(fields[-1]))
        else:
            results['hirshfeld_charges'].append(float(fields[-1]))
    return {k: np.array(v) for k, v in results.items()}


def _set_energy_and_free_energy(results: Dict[str, Any]):
    """Set values referred to as `energy` and `free_energy`."""
    if 'energy_with_dispersion_correction' in results:
        suffix = '_with_dispersion_correction'
    else:
        suffix = '_without_dispersion_correction'

    if 'free_energy' + suffix in results:  # metallic
        keye = 'energy_zero' + suffix
        keyf = 'free_energy' + suffix
    else:  # non-metallic
        # The finite basis set correction is applied to the energy at finite T
        # (not the energy at 0 K). We should hence refer to the corrected value
        # as `energy` only when the free energy is unavailable, i.e., only when
        # FIX_OCCUPANCY : TRUE and thus no smearing is applied.
        if 'energy_with_finite_basis_set_correction' in results:
            keye = 'energy_with_finite_basis_set_correction'
        else:
            keye = 'energy' + suffix
        keyf = 'energy' + suffix

    results['energy'] = results[keye]
    results['free_energy'] = results[keyf]



import importlib
from typing import List, Tuple

from ase.utils import (
    get_python_package_path_description,
    search_current_git_hash,
)


def format_dependency(modname: str) -> Tuple[str, str]:
    """Return (name, info) for given module.

    If possible, info is the path to the module's package."""
    try:
        module = importlib.import_module(modname)
    except ImportError:
        return modname, 'not installed'

    version = getattr(module, '__version__', '?')
    name = f'{modname}-{version}'
    if modname == 'ase':
        githash = search_current_git_hash(module)
        if githash:
            name += f'-{githash:.10}'

    # (only packages have __path__, but we are importing packages.)
    info = get_python_package_path_description(module)
    return name, info


def all_dependencies() -> List[Tuple[str, str]]:
    names = ['ase', 'numpy', 'scipy', 'matplotlib', 'spglib',
             'ase_ext', 'flask', 'psycopg2', 'pyamg']
    return [format_dependency(name) for name in names]



"""
Execution of turbomole binaries and scripts:
define, dscf, grad, ridft, rdgrad, aoforce, jobex, NumForce
"""
import os
from subprocess import PIPE, Popen


def get_output_filename(basename):
    """return the output file name from the basename of the executable"""
    return 'ASE.TM.' + basename + '.out'


def check_bad_output(stderr):
    """check status written in stderr by turbomole executables"""
    if 'abnormally' in stderr or 'ended normally' not in stderr:
        raise OSError(f'Turbomole error: {stderr}')


def execute(args, input_str=''):
    """executes a turbomole executable and process the outputs"""

    stdout_file = get_output_filename(os.path.basename(args[0]))
    with open(stdout_file, 'w') as stdout:
        proc = Popen(args, stdin=PIPE, stderr=PIPE, stdout=stdout,
                     encoding='ASCII')
        _stdout_txt, stderr_txt = proc.communicate(input=input_str)
        check_bad_output(stderr_txt)
    return stdout_file



"""Test reading/writing in ASE on pathlib objects"""

import io
from pathlib import Path

import pytest

from ase.build import molecule
from ase.io import read, write
from ase.utils import PurePath, convert_string_to_fd, reader, writer

# Test reader/writer
teststr = 'Teststring!'


@writer
def mywrite(file, fdcmp=None):
    assert isinstance(file, io.TextIOBase)
    assert file.mode == 'w'

    # Print something to the file
    print(teststr, file=file)

    # Check that we didn't change the fd
    if fdcmp:
        assert file is fdcmp


@reader
def myread(file, fdcmp=None):
    assert isinstance(file, io.TextIOBase)
    assert file.mode == 'r'

    # Ensure we can read from file
    line = next(file)
    assert line.strip() == teststr

    # Check that we didn't change the fd
    if fdcmp:
        assert file is fdcmp


def test_pathlib_support(testdir):
    path = Path('tmp_plib_testdir')

    # Test PurePath catches path
    assert isinstance(path, PurePath)

    path.mkdir(exist_ok=True)

    myf = path / 'test.txt'

    with pytest.warns(FutureWarning):
        fd = convert_string_to_fd(myf)
        fd.close()
        assert isinstance(fd, io.TextIOBase)

    with pytest.warns(FutureWarning):
        fd = convert_string_to_fd(str(myf))
        fd.close()
        assert isinstance(fd, io.TextIOBase)

    for f in [myf, str(myf)]:
        myf.unlink()                # Remove the file first
        mywrite(f)
        myread(f)

    # Check reader, writer on open filestream
    # Here, the filestream shouldn't be altered
    with myf.open('w') as fd:
        mywrite(fd, fdcmp=fd)
    with myf.open('r') as fd:
        myread(fd, fdcmp=fd)

    # Check that we can read and write atoms with pathlib
    atoms = molecule('H2', vacuum=5)
    f2 = path / 'test2.txt'
    for form in ['vasp', 'traj', 'xyz']:
        write(f2, atoms, format=form)
        read(f2, format=form)



import numpy as np


class Prior():
    """Base class for all priors for the bayesian optimizer.

    The __init__ method and the prior method are implemented here.
    Each child class should implement its own potential method, that will be
    called by the prior method implemented here.

    When used, the prior should be initialized outside the optimizer and the
    Prior object should be passed as a function to the optimizer.
    """

    def __init__(self):
        """Basic prior implementation."""

    def prior(self, x):
        """Actual prior function, common to all Priors"""
        if len(x.shape) > 1:
            n = x.shape[0]
            return np.hstack([self.potential(x[i, :]) for i in range(n)])
        else:
            return self.potential(x)


class ZeroPrior(Prior):
    """ZeroPrior object, consisting on a constant prior with 0eV energy."""

    def __init__(self):
        Prior.__init__(self)

    def potential(self, x):
        return np.zeros(x.shape[0] + 1)


class ConstantPrior(Prior):
    """Constant prior, with energy = constant and zero forces

    Parameters:

    constant: energy value for the constant.

    Example:

    >>> from ase.optimize import GPMin
    >>> from ase.optimize.gpmin.prior import ConstantPrior
    >>> op = GPMin(atoms, Prior = ConstantPrior(10)
    """

    def __init__(self, constant):
        self.constant = constant
        Prior.__init__(self)

    def potential(self, x):
        d = x.shape[0]
        output = np.zeros(d + 1)
        output[0] = self.constant
        return output

    def set_constant(self, constant):
        self.constant = constant


class CalculatorPrior(Prior):
    """CalculatorPrior object, allows the user to
    use another calculator as prior function instead of the
    default constant.

    Parameters:

    atoms: the Atoms object
    calculator: one of ASE's calculators
    """

    def __init__(self, atoms, calculator):
        Prior.__init__(self)
        self.atoms = atoms.copy()
        self.atoms.calc = calculator

    def potential(self, x):
        self.atoms.set_positions(x.reshape(-1, 3))
        V = self.atoms.get_potential_energy(force_consistent=True)
        gradV = -self.atoms.get_forces().reshape(-1)
        return np.append(np.array(V).reshape(-1), gradV)



import os
from io import StringIO

import ase.io as aio
from ase.io.gromacs import read_gromacs, write_gromacs


def test_gromacs_1():
    standard_gromacs = '''MD of 2 waters, t= 0.0
    6
    1WATER  OW1    1   0.126   1.624   1.679  0.1227 -0.0580  0.0434
    1WATER  HW2    2   0.190   1.661   1.747  0.8085  0.3191 -0.7791
    1WATER  HW3    3   0.177   1.568   1.613 -0.9045 -2.6469  1.3180
    2WATER  OW1    4   1.275   0.053   0.622  0.2519  0.3140 -0.1734
    2WATER  HW2    5   1.337   0.002   0.680 -1.0641 -1.1349  0.0257
    2WATER  HW3    6   1.326   0.120   0.568  1.9427 -0.8216 -0.0244
   1.82060   1.82060   1.82060
'''
    # read x from StringIO
    gromacs_file = StringIO(standard_gromacs)
    x = read_gromacs(gromacs_file)

    # write x to StringIO
    out_stringio = StringIO()
    write_gromacs(out_stringio, x)
    out_stringio.seek(0)

    # verify
    output = out_stringio.read()
    ase_output = '\n'.join(output.split('\n')[1:-2])
    standard_output = '\n'.join(standard_gromacs.split('\n')[1:-2])
    # import pdb; pdb.set_trace()
    assert (ase_output == standard_output
            ), 'ASE gromacs output differs from standard output'


def test_gromacs_2(tmp_path: str):
    water_str = """water
    3
    1SOL  OW000    1   0.000   0.000   0.000
    1SOL  HW001    2   0.000   0.100   0.000
    1SOL  HW002    3   0.088  -0.036   0.000
   1.00000   1.00000   1.00000\n"""
    gro_file = os.path.join(tmp_path, "water.gro")
    with open(gro_file, "w") as f:
        f.write(water_str)

    ase_atoms = read_gromacs(gro_file)
    assert str(ase_atoms.symbols) != "W3"

    byte_atoms = read_gromacs(gro_file)
    assert str(byte_atoms.symbols) == "OH2"

    dump_gro_file = os.path.join(tmp_path, "dump_water.gro")
    aio.write(dump_gro_file, byte_atoms)
    reload_atoms = read_gromacs(dump_gro_file)
    for atom_a, atom_b in zip(byte_atoms, reload_atoms):
        assert atom_a.symbol == atom_b.symbol
        assert all(atom_a.position == atom_b.position)



import pytest

from ase.lattice.cubic import FaceCenteredCubic


@pytest.fixture()
def lattice_params():
    lattice_params = {}
    lattice_params["size"] = (2, 2, 2)
    lattice_params["latticeconstant"] = 3.52
    lattice_params["symbol"] = "Ni"
    lattice_params["pbc"] = True
    return lattice_params


@pytest.mark.calculator_lite()
@pytest.mark.calculator("lammpslib")
def test_lammpslib_change_cell_bcs(factory, lattice_params, calc_params_NiH):
    """Test that a change in unit cell boundary conditions is
    handled correctly by lammpslib"""

    atoms = FaceCenteredCubic(**lattice_params)

    calc = factory.calc(**calc_params_NiH)
    atoms.calc = calc

    energy_ppp_ref = -142.400000403
    energy_ppp = atoms.get_potential_energy()
    print(f"Computed energy with boundary ppp = {energy_ppp}")
    assert energy_ppp == pytest.approx(energy_ppp_ref, rel=1e-4)

    atoms.set_pbc((False, False, True))
    energy_ssp_ref = -114.524625705
    energy_ssp = atoms.get_potential_energy()
    print(f"Computed energy with boundary ssp = {energy_ssp}")
    assert energy_ssp == pytest.approx(energy_ssp_ref, rel=1e-4)



# type: ignore
import io
import os
import unittest

import numpy as np
import pytest

import ase
import ase.build
import ase.io
from ase.build import graphene_nanoribbon
from ase.calculators.calculator import compare_atoms
from ase.constraints import (
    FixAtoms,
    FixedLine,
    FixedPlane,
    FixScaled,
    constrained_indices,
)
from ase.io.vasp import read_vasp_xdatcar, write_vasp_xdatcar


class TestXdatcarRoundtrip(unittest.TestCase):
    def setUp(self):
        self.outfile = 'NaCl.XDATCAR'

        self.NaCl = ase.build.bulk('NaCl', 'rocksalt', a=5.64)

    def tearDown(self):
        if os.path.isfile(self.outfile):
            os.remove(self.outfile)

    def assert_atoms_almost_equal(self, atoms, other, tol=1e-15):
        """Compare two Atoms objects, raising AssertionError if different"""
        system_changes = compare_atoms(atoms, other, tol=tol)

        if len(system_changes) > 0:
            raise AssertionError(
                'Atoms objects differ by {}'.format(', '.join(system_changes))
            )

    def assert_trajectory_almost_equal(self, traj1, traj2):
        self.assertEqual(len(traj1), len(traj2))
        for image, other in zip(traj1, traj2):
            self.assert_atoms_almost_equal(image, other)

    def test_roundtrip(self):
        # Create a series of translated cells
        trajectory = [self.NaCl.copy() for _ in range(5)]
        for i, atoms in enumerate(trajectory):
            atoms.set_scaled_positions(
                atoms.get_scaled_positions() + i * np.array([0.05, 0, 0.02])
            )
            atoms.wrap()

        ase.io.write(self.outfile, trajectory, format='vasp-xdatcar')
        roundtrip_trajectory = ase.io.read(self.outfile, index=':')
        self.assert_trajectory_almost_equal(trajectory, roundtrip_trajectory)

    def test_roundtrip_single_atoms(self):
        atoms = ase.build.bulk('Ge')
        ase.io.write(self.outfile, atoms, format='vasp-xdatcar')
        roundtrip_atoms = ase.io.read(self.outfile)
        self.assert_atoms_almost_equal(atoms, roundtrip_atoms)

    def test_typeerror(self):
        with self.assertRaises(TypeError):
            atoms = ase.build.bulk('Ge')
            write_vasp_xdatcar(self.outfile, atoms)
        with self.assertRaises(TypeError):
            not_atoms = 1
            ase.io.write(self.outfile, not_atoms, format='vasp-xdatcar')
        with self.assertRaises(TypeError):
            not_traj = [True, False, False]
            ase.io.write(self.outfile, not_traj, format='vasp-xdatcar')


def test_index():
    """Test if the `index` option works correctly"""
    atoms0 = ase.build.bulk('X', 'sc', a=1.0)
    atoms1 = atoms0.copy()
    atoms1.positions += 0.1
    images = (atoms0, atoms1)
    with io.StringIO() as buf:
        write_vasp_xdatcar(buf, images)

        buf.seek(0)
        atoms = read_vasp_xdatcar(buf, index=0)  # atoms0
        np.testing.assert_allclose(atoms.positions, [[0.0, 0.0, 0.0]])

        buf.seek(0)
        atoms = read_vasp_xdatcar(buf, index=-1)  # atoms1
        np.testing.assert_allclose(atoms.positions, [[0.1, 0.1, 0.1]])

        buf.seek(0)
        images = read_vasp_xdatcar(buf, index=None)  # atoms1
        np.testing.assert_allclose(atoms.positions, [[0.1, 0.1, 0.1]])

        buf.seek(0)
        images = read_vasp_xdatcar(buf, index=':')  # (atoms0, atoms1)
        assert isinstance(images, list)


# Start of tests for constraints
# VASP supports FixAtoms and FixScaled as well as FixedLine and
# FixedPlane if the direction is along a lattice vector. Test that
# these constraints are preserved when writing and reading POSCAR
# files.
indices_to_constrain = [0, 2]


@pytest.fixture()
def graphene_atoms():
    atoms = graphene_nanoribbon(2, 2, type="armchair", saturated=False)
    atoms.cell = [[10, 0, 0], [0, 10, 0], [0, 0, 10]]
    return atoms


def poscar_roundtrip(atoms):
    """Write a POSCAR file, read it back and return the new atoms object"""
    atoms.write("POSCAR", direct=True)
    return ase.io.read("POSCAR")


def test_FixAtoms(graphene_atoms):
    atoms = graphene_atoms
    atoms.set_constraint(FixAtoms(indices=indices_to_constrain))
    new_atoms = poscar_roundtrip(atoms)

    # Assert that constraints are preserved
    assert isinstance(new_atoms.constraints[0], FixAtoms)
    assert np.all(new_atoms.constraints[0].index == indices_to_constrain)


def test_FixScaled(graphene_atoms):
    atoms = graphene_atoms
    atoms.set_constraint(FixScaled(indices_to_constrain, mask=[0, 1, 1]))
    new_atoms = poscar_roundtrip(atoms)

    # Assert that constraints are preserved
    assert np.all(constrained_indices(new_atoms) == indices_to_constrain)
    assert np.all(new_atoms.constraints[0].mask == [0, 1, 1])


@pytest.mark.parametrize('ConstraintClass', [FixedLine, FixedPlane])
def test_FixedLine_and_Plane(ConstraintClass, graphene_atoms):
    atoms = graphene_atoms
    atoms.set_constraint(
        ConstraintClass(indices=indices_to_constrain, direction=[1, 0, 0])
    )
    new_atoms = poscar_roundtrip(atoms)

    # FixedLine and FixedPlane are converted to FixScaled. During
    # a relaxation the results will be the same since FixScaled
    # are equivalent to the others if the direction in FixedLine
    # or FixedPlane is along a lattice vector.

    assert np.all(constrained_indices(new_atoms) == indices_to_constrain)



from io import BytesIO

from ase.io import iread, write


def to_bytes(images, format=None, **kwargs):
    """Convert atoms or multiple atoms objects to bytes."""
    buf = _to_buffer(images, format=format, **kwargs)
    btxt = buf.getvalue()
    return btxt


def _to_buffer(images, format=None, **kwargs):
    buf = BytesIO()
    write(buf, images, format=format, **kwargs)
    buf.seek(0)
    return buf


def parse_images(data, format=None, **kwargs):
    """Parse string or bytes into list of atoms objects."""
    buf = BytesIO(data)
    images = list(iread(buf, format=format, **kwargs))
    return images


def parse_atoms(data, format=None, **kwargs):
    images = parse_images(data, format=format, index=-1, **kwargs)
    assert len(images) == 1
    return images[0]



"""Temporary file while we deprecate this locaation."""

from ase.mep import DimerControl as RealDimerControl
from ase.mep import MinModeAtoms as RealMinModeAtoms
from ase.mep import MinModeTranslate as RealMinModeTranslate
from ase.utils import deprecated


class DimerControl(RealDimerControl):
    @deprecated('Please import DimerControl from ase.mep, not ase.dimer.')
    def __init__(self, *args, **kwargs):
        """
        .. deprecated:: 3.23.0
            Please import ``DimerControl`` from :mod:`ase.mep`
        """
        super().__init__(*args, **kwargs)


class MinModeAtoms(RealMinModeAtoms):
    @deprecated('Please import MinModeAtoms from ase.mep, not ase.dimer.')
    def __init__(self, *args, **kwargs):
        """
        .. deprecated:: 3.23.0
            Please import ``MinModeAtoms`` from :mod:`ase.mep`
        """
        super().__init__(*args, **kwargs)


class MinModeTranslate(RealMinModeTranslate):
    @deprecated('Please import MinModeTranslate from ase.mep, not ase.dimer.')
    def __init__(self, *args, **kwargs):
        """
        .. deprecated:: 3.23.0
            Please import ``MinModeTranslate`` from :mod:`ase.mep`
        """
        super().__init__(*args, **kwargs)



import pytest

from ase.build import molecule
from ase.utils import workdir

calc = pytest.mark.calculator
filterwarnings = pytest.mark.filterwarnings


def _calculate(code, name):
    atoms = molecule(name)
    atoms.center(vacuum=3.5)
    with workdir(f'test-{name}', mkdir=True):
        atoms.calc = code.calc()
        return atoms.get_potential_energy()


omx_par = {'definition_of_atomic_species': [['C', 'C6.0', 'C_CA19'],
                                            ['H', 'H5.0', 'H_CA19']]}


@pytest.mark.calculator_lite()
@calc('abinit', ecut=300, chksymbreak=0, toldfe=1e-4)
@calc('aims')
@calc('cp2k')
@calc('espresso')
@calc('gpaw', symmetry='off', mode='pw', txt='gpaw.txt', mixer={'beta': 0.6},
      marks=[filterwarnings('ignore:.*?ignore_bad_restart_file'),
             filterwarnings('ignore:convert_string_to_fd')])
@calc('nwchem')
@calc('octopus', Spacing='0.25 * angstrom', BoxShape='minimum',
      convreldens=1e-3, Radius='3.5 * angstrom')
@calc('openmx', **omx_par)
@calc('siesta', marks=pytest.mark.xfail)
@calc('gamess_us', label='ch4')
@calc('gaussian', xc='lda', basis='3-21G')
def test_ch4_reaction(factory):
    e_ch4 = _calculate(factory, 'CH4')
    e_c2h2 = _calculate(factory, 'C2H2')
    e_h2 = _calculate(factory, 'H2')
    energy = e_ch4 - 0.5 * e_c2h2 - 1.5 * e_h2
    print(energy)
    ref_energy = -2.8
    assert abs(energy - ref_energy) < 0.3



from ase.atoms import Atoms
from ase.constraints import FixScaled


def test_negativeindex():
    a1 = Atoms(symbols='X2',
               positions=[[0., 0., 0.], [2., 0., 0.]],
               cell=[[4., 0., 0.], [0., 4., 0.], [0., 0., 4.]])

    fs1 = FixScaled(-1, mask=(True, False, False))
    fs2 = FixScaled(1, mask=(False, True, False))

    a1.set_constraint([fs1, fs2])

    # reassigning using atoms.__getitem__
    a2 = a1[0:2]

    assert len(a1.constraints) == len(a2.constraints)



import numpy as np


class GreenFunction:
    """Equilibrium retarded Green function."""

    def __init__(self, H, S=None, selfenergies=[], eta=1e-4):
        self.H = H
        self.S = S
        self.selfenergies = selfenergies
        self.eta = eta
        self.energy = None
        self.Ginv = np.empty(H.shape, complex)

    def retarded(self, energy, inverse=False):
        """Get retarded Green function at specified energy.

        If 'inverse' is True, the inverse Green function is returned (faster).
        """
        if energy != self.energy:
            self.energy = energy
            z = energy + self.eta * 1.j

            if self.S is None:
                self.Ginv[:] = 0.0
                self.Ginv.flat[:: len(self.S) + 1] = z
            else:
                self.Ginv[:] = z
                self.Ginv *= self.S
            self.Ginv -= self.H

            for selfenergy in self.selfenergies:
                self.Ginv -= selfenergy.retarded(energy)

        if inverse:
            return self.Ginv
        else:
            return np.linalg.inv(self.Ginv)

    def calculate(self, energy, sigma):
        """XXX is this really needed"""
        ginv = energy * self.S - self.H - sigma
        return np.linalg.inv(ginv)

    def apply_retarded(self, energy, X):
        """Apply retarded Green function to X.

        Returns the matrix product G^r(e) . X
        """
        return np.linalg.solve(self.retarded(energy, inverse=True), X)

    def dos(self, energy):
        """Total density of states -1/pi Im(Tr(GS))"""
        if self.S is None:
            return -self.retarded(energy).imag.trace() / np.pi
        else:
            GS = self.apply_retarded(energy, self.S)
            return -GS.imag.trace() / np.pi

    def pdos(self, energy):
        """Projected density of states -1/pi Im(SGS/S)"""
        if self.S is None:
            return -self.retarded(energy).imag.diagonal() / np.pi
        else:
            S = self.S
            SGS = np.dot(S, self.apply_retarded(energy, S))
            return -(SGS.diagonal() / S.diagonal()).imag / np.pi



import numpy as np
import pytest

from ase.lattice import FCC
from ase.spacegroup import Spacegroup
from ase.spacegroup.spacegroup import (
    parse_sitesym,
    parse_sitesym_element,
    parse_sitesym_single,
)


def test_spacegroup_miscellaneous():
    no = 225
    sg = Spacegroup(no)
    assert int(sg) == no == sg.no
    assert sg.centrosymmetric
    assert sg.symbol == 'F m -3 m'
    assert sg.symbol in str(sg)
    assert sg.lattice == 'F'  # face-centered
    assert sg.setting == 1
    assert sg.scaled_primitive_cell == pytest.approx(FCC(1.0).tocell()[:])
    assert sg.reciprocal_cell @ sg.scaled_primitive_cell == pytest.approx(
        np.identity(3))


def test_spacegroup_parse_sitesym():
    sitesyms = ["x,y,z", "0.25-y,x+1/4,-z",
                "x-1/4, y+3/4,   1/8-z", "x-y,-z+0.3-y,-1/2+x-y+z"]
    expected_trans = np.array(
        [
            [+0.000, 0.000, 0.000],  # x, y, z
            [+0.250, 0.250, 0.000],  # 0.25-y, x+1/4, -z
            [-0.250, 0.750, 0.125],   # x-1/4, y+3/4, 1/8-z
            [+0.000, 0.300, -0.500]  # x-y, -z+0.3-y, -1/2+x-y+z
        ],
        dtype=float
    )
    expected_trans_bounded = expected_trans.copy()
    expected_trans_bounded[2, 0] = 0.75
    expected_trans_bounded[3, 2] = 0.50
    expected_rot = np.array(
        [
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],  # x, y, z
            [[0, -1, 0], [1, 0, 0], [0, 0, -1]],  # 0.25-y, x+1/4, -z
            [[1, 0, 0], [0, 1, 0], [0, 0, -1]],  # x-1/4, y+3/4, 1/8-z
            [[1, -1, 0], [0, -1, -1], [1, -1, 1]]  # # x-y, -z+0.3-y, -1/2+x-y+z
        ],
        dtype=int
    )

    rot, trans = parse_sitesym(sitesyms)
    rot_bounded, trans_bounded = parse_sitesym(
        sitesyms, force_positive_translation=True)
    assert np.allclose(rot, expected_rot)
    assert np.allclose(rot_bounded, expected_rot)
    assert np.allclose(trans, expected_trans)
    assert np.allclose(trans_bounded, expected_trans_bounded)


def test_spacegroup_parse_sitesym_single():
    sitesym = "0.125-y,x-1/4,-z"
    expected_rot = np.array([[0, -1, 0], [1, 0, 0], [0, 0, -1]], dtype=int)
    expected_trans = np.array([0.125, -0.25, 0.0], dtype=float)
    rot = np.arange(9, dtype=int).reshape(3, 3)
    trans = np.zeros(3, dtype=float) + 2.0
    parse_sitesym_single(sitesym, out_rot=rot, out_trans=trans)
    assert np.allclose(rot, expected_rot)
    assert np.allclose(trans, expected_trans)


def test_spacegroup_parse_sitesym_element():
    element_list = ("x", "y", "z", "-x", "0.1+x", "-y-1/8", "+1/2-z", "x+z")
    expected_rot_list = ([(0, 1)], [(1, 1)], [(2, 1)],
                         [(0, -1)], [(0, 1)], [(1, -1)],
                         [(2, -1)], [(0, 1), (2, 1)])
    expected_trans_list = (0.0, 0.0, 0.0, 0.0, 0.1, -0.125, 0.5, 0.0)

    for element, expected_rot, expected_trans in zip(
            element_list, expected_rot_list, expected_trans_list):
        rot, trans = parse_sitesym_element(element)
        assert rot == expected_rot
        assert np.allclose(trans, expected_trans)



import pytest

from ase.build import bulk
from ase.io.siesta_output import OutputReader


def test_siesta_read_eigenvalues_soc(datadir, config_file, tmp_path):
    """ In this test, we read a stored siesta.EIG file."""
    reader = OutputReader(prefix='siesta', directory=tmp_path)
    assert reader.read_eigenvalues() == {}

    reader = OutputReader(prefix='siesta', directory=datadir / 'siesta')
    dct = reader.read_eigenvalues()
    assert dct['eigenvalues'].shape == (1, 1, 30)


@pytest.mark.calculator('siesta')
def test_siesta_read_eigenvalues(factory):
    # Test real calculation which produces a gapped .EIG file
    atoms = bulk('Si')
    calc = factory.calc(kpts=[2, 1, 1])
    atoms.calc = calc
    atoms.get_potential_energy()

    assert calc.results['eigenvalues'].shape[:2] == (1, 2)  # spins x kpts
    assert calc.get_k_point_weights().shape == (2,)
    assert calc.get_ibz_k_points().shape == (2, 3)
    assert calc.get_number_of_spins() == 1



"""colors.py - select how to color the atoms in the GUI."""
import numpy as np

import ase.gui.ui as ui
from ase.gui.i18n import _
from ase.gui.utils import get_magmoms


class ColorWindow:
    """A window for selecting how to color the atoms."""

    def __init__(self, gui):
        self.reset(gui)

    def reset(self, gui):
        """create a new color window"""
        self.win = ui.Window(_('Colors'), wmtype='utility')
        self.gui = gui
        self.win.add(ui.Label(_('Choose how the atoms are colored:')))
        values = ['jmol', 'tag', 'force', 'velocity',
                  'initial charge', 'magmom', 'neighbors']
        labels = [_('By atomic number, default "jmol" colors'),
                  _('By tag'),
                  _('By force'),
                  _('By velocity'),
                  _('By initial charge'),
                  _('By magnetic moment'),
                  _('By number of neighbors'), ]

        haveit = ['numbers', 'positions', 'forces', 'momenta',
                  'initial_charges', 'initial_magmoms']
        for key in self.gui.atoms.arrays:
            if key not in haveit:
                values.append(key)
                labels.append(f'By user-defined "{key}"')

        self.radio = ui.RadioButtons(labels, values, self.toggle,
                                     vertical=True)
        self.radio.value = gui.colormode
        self.win.add(self.radio)
        self.activate()
        self.label = ui.Label()
        self.win.add(self.label)

        if hasattr(self, 'mnmx'):
            self.win.add(self.cmaps)
            self.win.add(self.mnmx)

    def change_mnmx(self, mn=None, mx=None):
        """change min and/or max values for colormap"""
        if mn:
            self.mnmx[1].value = mn
        if mx:
            self.mnmx[3].value = mx
        mn, mx = self.mnmx[1].value, self.mnmx[3].value
        colorscale, _, _ = self.gui.colormode_data
        self.gui.colormode_data = colorscale, mn, mx
        self.gui.draw()

    def activate(self):
        images = self.gui.images
        atoms = self.gui.atoms
        radio = self.radio
        radio['tag'].active = atoms.has('tags')

        # XXX not sure how to deal with some images having forces,
        # and other images not.  Same goes for below quantities
        F = images.get_forces(atoms)
        radio['force'].active = F is not None
        radio['velocity'].active = atoms.has('momenta')
        radio['initial charge'].active = atoms.has('initial_charges')
        radio['magmom'].active = get_magmoms(atoms).any()
        radio['neighbors'].active = True

    def toggle(self, value):
        self.gui.colormode = value
        if value == 'jmol' or value == 'neighbors':
            if hasattr(self, 'mnmx'):
                "delete the min max fields by creating a new window"
                del self.mnmx
                del self.cmaps
                self.win.close()
                self.reset(self.gui)
            text = ''
        else:
            scalars = np.ma.array([self.gui.get_color_scalars(i)
                                   for i in range(len(self.gui.images))])
            mn = np.min(scalars)
            mx = np.max(scalars)
            self.gui.colormode_data = None, mn, mx

            cmaps = ['default', 'old']
            try:
                import pylab as plt
                cmaps += [m for m in plt.cm.datad if not m.endswith("_r")]
            except ImportError:
                pass
            self.cmaps = [_('cmap:'),
                          ui.ComboBox(cmaps, cmaps, self.update_colormap),
                          _('N:'),
                          ui.SpinBox(26, 0, 100, 1, self.update_colormap)]
            self.update_colormap('default')

            try:
                unit = {'tag': '',
                        'force': 'eV/Ang',
                        'velocity': '(eV/amu)^(1/2)',
                        'charge': '|e|',
                        'initial charge': '|e|',
                        'magmom': 'μB'}[value]
            except KeyError:
                unit = ''
            text = ''

            rng = mx - mn  # XXX what are optimal allowed range and steps ?
            self.mnmx = [_('min:'),
                         ui.SpinBox(mn, mn - 10 * rng, mx + rng, rng / 10.,
                                    self.change_mnmx, width=20),
                         _('max:'),
                         ui.SpinBox(mx, mn - 10 * rng, mx + rng, rng / 10.,
                                    self.change_mnmx, width=20),
                         _(unit)]
            self.win.close()
            self.reset(self.gui)

        self.label.text = text
        self.radio.value = value
        self.gui.draw()
        return text  # for testing

    def notify_atoms_changed(self):
        "Called by gui object when the atoms have changed."
        self.activate()
        mode = self.gui.colormode
        if not self.radio[mode].active:
            mode = 'jmol'
        self.toggle(mode)

    def update_colormap(self, cmap=None, N=26):
        "Called by gui when colormap has changed"
        import matplotlib
        if cmap is None:
            cmap = self.cmaps[1].value
        try:
            N = int(self.cmaps[3].value)
        except AttributeError:
            N = 26
        colorscale, mn, mx = self.gui.colormode_data
        if cmap == 'default':
            colorscale = ['#{0:02X}80{0:02X}'.format(int(red))
                          for red in np.linspace(0, 250, N)]
        elif cmap == 'old':
            colorscale = [f'#{int(red):02X}AA00'
                          for red in np.linspace(0, 230, N)]
        else:
            cmap_obj = matplotlib.colormaps[cmap]
            colorscale = [matplotlib.colors.rgb2hex(c[:3]) for c in
                          cmap_obj(np.linspace(0, 1, N))]
        self.gui.colormode_data = colorscale, mn, mx
        self.gui.draw()
